# ./workflow/__init__.py

from .step1_fastq_qc import step1_fastq_qc as step1_fastq_qc_main
from .step2_remove_host_reads import step2_remove_host_reads as step2_remove_host_reads_main

from .step2_remove_host_reads2 import step2_remove_host_reads2 as step2_remove_host_reads2_main

from .step3_kraken2 import step3_kraken2 as step3_kraken2_main

from .step4_extract_microbiome_reads import step4_extract_microbiome_reads as step4_extract_microbiome_reads_main

from .step5_decontamination import step5_decontamination as step5_decontamination_main

from .step6_kraken2Bracken import step6_kraken2Bracken as step6_kraken2Bracken_main

from .step7_merge_itm_results import step7_merge_itm_results as step7_merge_itm_results_main

from .step8_get_data_summary import step8_get_data_summary as step8_get_data_summary_main



import os
import argparse
from multiprocessing import Pool
from itmfinder.itm_helper.extract_kraken_reads import extract_reads_i

def print_colorful_message(message, color):
    """
    Print a colorful message to the console.

    Args:
        message (str): The message to be printed.
        color (str): The color code to be applied.
                     Choices include 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white'.
    """
    colors = {
        'red': '\033[91m',
        'green': '\033[92m',
        'yellow': '\033[93m',
        'blue': '\033[94m',
        'magenta': '\033[95m',
        'cyan': '\033[96m',
        'white': '\033[97m',
    }
    end_color = '\033[0m'
    if color not in colors:
        print("Invalid color. Please choose from 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white'.")
        return
    print(colors[color] + message + end_color)

def compress_to_gz(file_path):
    """
    Compress the given file to a .gz file and remove the original.

    Args:
        file_path (str): Path to the file to be compressed.
    """
    import gzip
    if os.path.exists(file_path):
        with open(file_path, 'rb') as f_in:
            with gzip.open(file_path + '.gz', 'wb') as f_out:
                f_out.writelines(f_in)
        os.remove(file_path)
        print(f"Compressed and removed original: {file_path}")
    else:
        print(f"File {file_path} not found. Compression skipped.")

def extract_microbiome_reads(sample_info, is_single_end=False, TID=None):
    """
    Extract microbiome reads for a given sample using the itmfinder extract_reads_main function.

    Args:
        sample_info (tuple): Contains information about the sample.
        is_single_end (bool): Indicates if the data is single-end (True) or paired-end (False).
        TID (list of int): Taxonomic IDs to extract reads for.
    """
    sample_id, fastq_r1, kraken_output, output_r1, path4_ku1, path5_mr = sample_info
    report_file = os.path.join(path4_ku1, f"{sample_id}.kraken.report.txt")
    task_complete_file = os.path.join(path5_mr, f"{sample_id}.task.complete")

    if os.path.exists(task_complete_file):
        print(f">>> Sample {sample_id} already processed. Skipping...")
        return

    if not TID:
        TID = [2, 10239, 4751, 2157]  # Default Taxonomic IDs

    if is_single_end:
        extract_reads_i(kraken_file=kraken_output, seq_file1=fastq_r1, output_file1=output_r1, taxid=TID, report_file=report_file)
    else:
        fastq_r2 = fastq_r1.replace("_R1", "_R2")
        output_r2 = output_r1.replace("_1", "_2")
        extract_reads_i(kraken_file=kraken_output, seq_file1=fastq_r1, seq_file2=fastq_r2, output_file1=output_r1, output_file2=output_r2, taxid=TID, report_file=report_file)

    # Mark the task as complete
    with open(task_complete_file, "w") as f:
        f.write("Task completed.")

    # Optionally compress the output files
    compress_to_gz(output_r1)
    if not is_single_end:
        compress_to_gz(output_r2)

    print(f">>>--- Sample {sample_id} processed successfully.")

def step4_extract_microbiome_reads(path3_hr, path4_ku1, path5_mr, batch_size=1, is_single_end=False, TID=None):
    """
    Process all samples to extract microbiome reads based on Kraken2 output.

    Args:
        path3_hr (str): Directory containing host-removed FASTQ files.
        path4_ku1 (str): Directory containing Kraken2 output files.
        path5_mr (str): Directory where extracted microbiome reads will be stored.
        batch_size (int): Number of samples to process in parallel.
        is_single_end (bool): Flag to indicate if the processing should be for single-end reads.
        TID (list of int): List of Taxonomic IDs to filter reads.
    """
    print("   ")
    print_colorful_message("#########################################################", "blue")
    print_colorful_message(" ITMfinder: Identifing Intratumoral Microbiome pipeline ", "cyan")
    print_colorful_message(" If you encounter any issues, please report them at ", "cyan")
    print_colorful_message(" https://github.com/LiaoWJLab/ITMtools/issues ", "cyan")
    print_colorful_message("#########################################################", "blue")
    print(" Author: Dongqiang Zeng, Qianqian Mao ")
    print(" Email: interlaken@smu.edu.cn ")
    print_colorful_message("#########################################################", "blue")
    print("   ")
    
    print(" >>> Step-4 Extract Microbiome Reads...  ")

    os.makedirs(path5_mr, exist_ok=True)

    fastq_files = [f for f in os.listdir(path3_hr) if f.endswith("_R1.fastq.gz")] if not is_single_end else [f for f in os.listdir(path3_hr) if f.endswith(".fastq.gz")]
    samples = [
        (f.replace("_R1.fastq.gz", "") if not is_single_end else f.replace(".fastq.gz", ""),
         os.path.join(path3_hr, f),
         os.path.join(path4_ku1, f"{f.replace('_R1.fastq.gz', '') if not is_single_end else f.replace('.fastq.gz', '')}.kraken.output.txt"),
         os.path.join(path5_mr, f"{f.replace('_R1.fastq.gz', '_mr_1.fastq')}" if not is_single_end else f.replace(".fastq.gz", "_mr.fastq")),
         path4_ku1, path5_mr) for f in fastq_files
    ]

    with Pool(processes=batch_size) as pool:
        pool.starmap(extract_microbiome_reads, [(sample, is_single_end, TID) for sample in samples])

    print(">>> Microbiome reads extraction completed.")

def main():
    parser = argparse.ArgumentParser(description="Step 4: Extract microbiome reads from Kraken2 output")
    parser.add_argument("--path3_hr", type=str, required=True, help="Path to host removed FASTQ files")
    parser.add_argument("--path4_ku1", type=str, required=True, help="Path to Kraken2 outputs")
    parser.add_argument("--path5_mr", type=str, required=True, help="Path to extracted microbiome reads")
    parser.add_argument("--batch_size", type=int, default=1, help="Number of samples to process simultaneously")
    parser.add_argument("--se", action='store_true', help="Use single-end processing. Default is paired-end.")
    parser.add_argument("--TID", nargs='*', type=int, help="Taxonomic IDs for filtering reads.")
    args = parser.parse_args()

    step4_extract_microbiome_reads(args.path3_hr, args.path4_ku1, args.path5_mr, args.batch_size, args.se, args.TID)

# Ensure the main function only runs if this file is executed directly
if __name__ == "__main__":
    main()





import os
import subprocess
import random
import argparse
from itmfinder.itm_helper.kreport2mpa import kreport2mpai

def print_colorful_message(message, color):
    """
    Print a colorful message to the console.

    Args:
        message (str): The message to be printed.
        color (str): The color code to be applied. 
                     'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white'
    """
    colors = {
        'red': '\033[91m',
        'green': '\033[92m',
        'yellow': '\033[93m',
        'blue': '\033[94m',
        'magenta': '\033[95m',
        'cyan': '\033[96m',
        'white': '\033[97m',
    }
    end_color = '\033[0m'
    if color not in colors:
        print("Invalid color. Please choose from 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white'.")
        return
    colored_message = f"{colors[color]}{message}{end_color}"
    print(colored_message)


def step3_kraken2(path3_hr, path4_ku1, db_ku, num_threads=8, se=False):
    """
    Perform taxonomic classification using Kraken2.

    Args:
        path3_hr (str): Path to host-removed FASTQ files.
        path4_ku1 (str): Path to Kraken2 outputs.
        db_ku (str): Path to Kraken2 database.
        num_threads (int): Number of threads for Kraken2 analysis. Default is 8.
        se (bool): Whether the data is single-end sequencing. Default is False.
    """
    print("   ")
    print_colorful_message("#########################################################", "blue")
    print_colorful_message(" ITMfinder: Identifing Intratumoral Microbiome pipeline ", "cyan")
    print_colorful_message(" If you encounter any issues, please report them at ", "cyan")
    print_colorful_message(" https://github.com/LiaoWJLab/ITMtools/issues ", "cyan")
    print_colorful_message("#########################################################", "blue")
    print(" Author: Dongqiang Zeng, Qianqian Mao ")
    print(" Email: interlaken@smu.edu.cn ")
    print_colorful_message("#########################################################", "blue")
    print("   ")
    
    print(" >>> Step-3 Kraken2 Classification...  ")

    # Create the output directory if it does not exist
    os.makedirs(path4_ku1, exist_ok=True)
    
    # List files in the host-removed FASTQ directory
    files = os.listdir(path3_hr)
    files = [file for file in files if file.endswith("_host_remove_R1.fastq.gz")]
    random.shuffle(files)  # Shuffle the files randomly
    total_files = len(files)
    current_file = 0
    
    for file in files:
        current_file += 1
        print(f">>> Processing file {current_file} of {total_files}: {file}")
        # Extract sample ID from the filename
        sample_id = file[:-len("_host_remove_R1.fastq.gz")]
        # Construct paths to input and output files
        input_fq1 = os.path.join(path3_hr, f"{sample_id}_host_remove_R1.fastq.gz")
        input_fq2 = os.path.join(path3_hr, f"{sample_id}_host_remove_R2.fastq.gz")
        output_file = os.path.join(path4_ku1, f"{sample_id}.kraken.output.txt")
        report = os.path.join(path4_ku1, f"{sample_id}.kraken.report.txt")
        report_std = os.path.join(path4_ku1, f"{sample_id}.kraken.report.std.txt")
        mpa_output = os.path.join(path4_ku1, f"{sample_id}.kraken.mpa.txt")
        task_complete = os.path.join(path4_ku1, f"{sample_id}.task.complete")
        # Check if the task is already completed
        if os.path.exists(task_complete) and os.path.getsize(report_std) > 0:
            print(f"{report_std} already exists, skipping...")
        else:
            # Run Kraken2 for taxonomic classification
            print(f">>>-- Running Kraken2 for taxonomic classification on sample {sample_id}...")
            if se:
                subprocess.run(["kraken2", "--db", db_ku, "--threads", str(num_threads), "--report-minimizer-data", "--report", report,
                                "--use-names", "--output", output_file, input_fq1])
            else:
                subprocess.run(["kraken2", "--db", db_ku, "--threads", str(num_threads), "--report-minimizer-data", "--report", report,
                                "--use-names", "--output", output_file, "--paired", input_fq1, input_fq2])
            # Process Kraken report to a more readable format
            print("      Processing Kraken report...")
            subprocess.run(f"cut -f1-3,6-8 {report} > {report_std}", shell=True)
            # Convert Kraken report to MPA format
            print("      Converting Kraken report to MPA format...")

            # Convert report to MPA format using directly imported function
            kreport2mpai(report_std, mpa_output)
            # subprocess.run(["kreport2mpai", "-r", report_std, "-o", mpa_output])
            # Create a task complete file
            open(task_complete, "w").close()
            print("   ")
    
    print(">>> Kraken2 taxonomic classification completed.")

def main():
    parser = argparse.ArgumentParser(description="Step 3: Taxonomic classification using Kraken2")
    parser.add_argument("--path3_hr", type=str, help="Path to host removed FASTQ files")
    parser.add_argument("--path4_ku1", type=str, help="Path to Kraken2 outputs")
    parser.add_argument("--db_ku", type=str, help="Path to Kraken2 database")
    parser.add_argument("--num_threads", type=int, default=8, help="Number of threads")
    parser.add_argument("--se", action='store_true', help="Single-end sequencing data")
    args = parser.parse_args()

    step3_kraken2(args.path3_hr, args.path4_ku1, args.db_ku, args.num_threads, args.se)

# Ensure the main function only runs if this file is executed directly
if __name__ == "__main__":
    main()




#! /usr/bin/env python 
#####################################################################
#combine_bracken_output.py combines multiple Bracken output files for comparison
#Copyright (C) 2016-2023 Jennifer Lu, jlu26@jhmi.edu

#This file is part of Bracken.
          
#Bracken is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 3 of the license, or
#(at your option) any later version.

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with this program; if not, see <http://www.gnu.org/licenses/>.

#####################################################################
#Jennifer Lu, jlu26@jhmi.edu
#02/21/2018
#
#This program takes multiple Bracken OUTPUT files and creates a single 
#text tab-delimined file with estimated abundance across all samples
#for comparison. 
#
#Input files:
#   - Bracken output files (not Bracken report files)
#       [NOTE: All samples being compared must have the same level estimated]
#
#Optional Input Parameter:
#   - Sample Names for each sample separated by commas.
#     Number of sample names must match the number of samples
#     If not given, basename of sample files will be used for column headers
#       [e.g. --names S1,S2 will yield column headers S1_num S1_frac S2_num S2_frac]
#
#Output File format (tab-delimited, for species abundance estimation)
#   - Name of species
#   - Taxonomy ID for that species
#   - Taxonomy Level specified for
#   - Sample #1: Number of reads estimated for that species  
#   - Sample #1: Fraction of total reads in the sample estimated for this species
#       [NOTE: Fractions are of total reads classified, unclassified reads not accounted for]
#   - Sample #2: Number of reads 
#   - Sample #2: Fraction of total reads
#   - ...etc.
#
#Methods:
#   - main
#####################################################################
import os, sys, argparse
from time import gmtime
from time import strftime 

#Main method
def main():
    #Parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('--files', dest='files', 
        nargs='+', type=str, required=True,
        help='Bracken output files to combine.')
    parser.add_argument('--names', dest='names', 
        default='',required=False,
        help='Names for each input file - to be used in column headers of output [separate names with commas]')
    parser.add_argument('-o', '--output', dest='output', required=True,
        help='Name of output file with combined Bracken results.')
    args = parser.parse_args()

    #Start program
    time = strftime("%m-%d-%Y %H:%M:%S", gmtime())
    sys.stdout.write("PROGRAM START TIME: " + time + '\n')

    #Initialize variables 
    sample_counts = {}  #species :: sample1: counts, samples2: counts 
    total_reads = {}    #sample1: totalcounts, sample2: totalcounts 
    all_samples = []
    #Get sample names
    if len(args.names) == 0:
        for f in args.files:
            curr_sample = os.path.basename(f)
            total_reads[curr_sample] = 0
            all_samples.append(curr_sample)
    else:
        for curr_sample in args.names.split(","):
            total_reads[curr_sample] = 0
            all_samples.append(curr_sample) 
    #Read each file information in 
    level = ''
    i = 0
    for f in args.files:
        #Print update
        curr_name = all_samples[i]
        i += 1
        sys.stdout.write("Processing Output File %s:: Sample %s\n" % (f, curr_name))
        #Iterate through file
        header = True
        i_file = open(f,'r')
        for line in i_file:
            #Header line
            if header:
                header=False
                continue
            #Process line 
            [name, taxid, taxlvl, kreads, areads, estreads, frac] = line.strip().split("\t")
            estreads = int(estreads) 
            #Error Checks
            if name not in sample_counts:
                sample_counts[name] = {}
                sample_counts[name][taxid] = {}
            elif taxid != list(sample_counts[name].keys())[0]:
                sys.exit("Taxonomy IDs not matching for species %s: (%s\t%s)" % (name, taxid, sample_counts[name].keys()[0]))
            if len(level) == 0:
                level = taxlvl 
            elif level != taxlvl:
                sys.exit("Taxonomy level not matching between samples")
            #Save counts
            total_reads[curr_name] += estreads
            sample_counts[name][taxid][curr_name] = estreads 
        #Close file 
        i_file.close()

    #Print output file header
    o_file = open(args.output, 'w')
    o_file.write("name\ttaxonomy_id\ttaxonomy_lvl")
    for name in all_samples:
        o_file.write("\t%s_num\t%s_frac" % (name,name))
    o_file.write("\n")
    #Print each sample 
    for name in sample_counts:
        #Print information for classification
        taxid = list(sample_counts[name].keys())[0]
        o_file.write("%s\t%s\t%s" % (name, taxid, level)) 
        #Calculate and print information per sample 
        for sample in all_samples:
            if sample in sample_counts[name][taxid]:
                num = sample_counts[name][taxid][sample]
                perc = float(num)/float(total_reads[sample])
                o_file.write("\t%i\t%0.5f" % (num,perc))
            else:
                o_file.write("\t0\t0.00000")
        o_file.write("\n")
    o_file.close()

    #End program
    time = strftime("%m-%d-%Y %H:%M:%S", gmtime())
    sys.stdout.write("PROGRAM END TIME: " + time + '\n')

if __name__ == "__main__":
    main()



import os
import pandas as pd
import glob
from itmfinder.itm_helper.kraken2otu import main as kraken2otu_main
from itmfinder.itm_helper.combine_bracken_outputs import main as combine_bracken_main
from itmfinder.itm_helper.combine_mpa import main as combine_mpa_main


def print_colorful_message(message, color):
    colors = {
        'red': '\033[91m',
        'green': '\033[92m',
        'blue': '\033[94m',
        'magenta': '\033[95m',
        'cyan': '\033[96m',
        'yellow': '\033[93m',
        'white': '\033[97m',
    }
    end_color = '\033[0m'
    colored_message = f"{colors.get(color, '')}{message}{end_color}"
    print(colored_message)

def run_combine_bracken_outputs(path8_bracken, path_out):
    bracken_files = glob.glob(os.path.join(path8_bracken, "*.bracken"))
    level_mapping = {
        'g': 'genus',
        'f': 'family',
        'o': 'order',
        's': 'species'
    }

    for level, level_name in level_mapping.items():
        level_files = [file for file in bracken_files if file.endswith(f"{level}.bracken")]
        
        if not level_files:
            print(f"No {level}.bracken files found.")
            continue

        output_file = os.path.join(path_out, f"2-combined_bracken_results_{level_name}.txt")
        
        combine_bracken_main(
            files=level_files,
            output=output_file,
            taxonomy_level=level_name
        )
        print(f"Combined Bracken results saved to: {output_file}")

def step7_merge_itm_results(path7_ku2, path9_mpa, path8_bracken, mpa_suffix=".kraken.mpa.std.txt", report_suffix="kraken.report.txt", output_file="1-combine_mpa_std.txt"):
    os.makedirs(path9_mpa, exist_ok=True)

    print("   ")
    print_colorful_message("#########################################################", "blue")
    print_colorful_message(" ITMfinder: Identifying Intratumoral Microbiome pipeline ", "cyan")
    print_colorful_message(" If you encounter any issues, please report them at ", "cyan")
    print_colorful_message(" https://github.com/LiaoWJLab/ITMfinder/issues ", "cyan")
    print_colorful_message("#########################################################", "blue")
    print(" Author: Dongqiang Zeng, Qianqian Mao ")
    print(" Email: interlaken@smu.edu.cn ")
    print_colorful_message("#########################################################", "blue")
    print("   ")
    
    print(" >>> Step-7 Merge Results to Matrix and MPA Formats...  ")

    # Combine MPA files
    mpa_files = [file for file in os.listdir(path7_ku2) if file.lower().endswith(mpa_suffix) and os.path.getsize(os.path.join(path7_ku2, file)) > 0]
    if mpa_files:
        mpa_output_path = os.path.join(path9_mpa, output_file)
        combine_mpa_main(
            mpa_input_dir=path7_ku2,
            combined_output=mpa_output_path,
            level="all"
        )
        update_mpa_column_names(mpa_output_path, mpa_files, mpa_suffix)

    # Generate OTU tables
    levels = ["c", "o", "f", "g", "s"]
    empty_files = check_empty_files(path7_ku2)
    if empty_files:
        print_colorful_message(f"Warning: The following files are empty and will be deleted: {empty_files}", "yellow")
        for empty_file in empty_files:
            os.remove(empty_file)

    for level in levels:
        kraken2otu_main(
            kraken_input=path7_ku2,
            level=level,
            extension=report_suffix,
            output=os.path.join(path9_mpa, f"otu_table_{level}.csv")
        )

    # Combine Bracken outputs
    run_combine_bracken_outputs(path8_bracken, path9_mpa)

def update_mpa_column_names(file_path, mpa_files, mpa_suffix):
    df = pd.read_csv(file_path, sep='\t')
    sample_ids = [file.replace(mpa_suffix, "") for file in mpa_files]
    updated_column_names = ["TaxonID"] + sample_ids

    if len(updated_column_names) != len(df.columns):
        print("Column count mismatch in MPA output file.")
        return

    df.columns = updated_column_names
    df.to_csv(file_path, sep='\t', index=False)
    print("Updated column names in MPA file.")
    print(df.head())

def check_empty_files(path):
    empty_files = []
    for file in os.listdir(path):
        file_path = os.path.join(path, file)
        if file.endswith("kraken.report.txt") and os.path.getsize(file_path) == 0:
            print(f">>>== Empty file found: {file_path}. This file will be deleted.")
            empty_files.append(file_path)
    return empty_files

def main():
    parser = argparse.ArgumentParser(description="Step 7: Merge data to OTU and MPA formats")
    parser.add_argument("--path7_ku2", type=str, help="Path to Kraken2 outputs for report")
    parser.add_argument("--path9_mpa", type=str, help="Path to merged OTU and MPA data")
    parser.add_argument("--path8_bracken", type=str, help="Path to Bracken output files")
    parser.add_argument("--mpa_suffix", type=str, default=".kraken.mpa.std.txt", help="Suffix for MPA files. Default is '.kraken.mpa.std.txt'.")
    parser.add_argument("--report_suffix", type=str, default="kraken.report.txt", help="Suffix for report files. Default is 'kraken.report.txt'.")
    parser.add_argument("--output_file", type=str, default="1-combine_mpa_std.txt", help="Name of the output file. Default is '1-combine_mpa_std.txt'.")
    args = parser.parse_args()

    step7_merge_itm_results(args.path7_ku2, args.path9_mpa, args.path8_bracken, args.mpa_suffix, args.report_suffix, args.output_file)

# Ensure the main function only runs if this file is executed directly
if __name__ == "__main__":
    main()




#!/usr/bin/env python
######################################################################
#make_kreport.py takes in the kraken output file and the make_ktaxonomy.py 
#output file to generate a kraken report file 
#Copyright (C) 2020 Jennifer Lu, jennifer.lu717@gmail.com
#
#This file is part of KrakenTools
#KrakenTools is free software; oyu can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 3 of the license, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of 
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program; if not, see <http://www.gnu.org/licenses/>.
#
######################################################################
#Jennifer Lu, jlu26@jhmi.edu
#Updated: 04/15/2020
#
#This program creates the kraken report file from
#the make_ktaxonomy.py output and the kraken output file
#
#Required Parameters:
#   -i,-k,--kraken X....................kraken output file
#   -t,--taxonomy X.....................taxonomy file 
#   -o, --output X......................output kraken report file 
#Optional Parameters:
#   -h, --help..........................show help message.
#################################################################################
import os, sys, argparse
import operator
from time import gmtime
from time import strftime 
#################################################################################
#Tree Class
#usage: tree node used in constructing taxonomy tree
class Tree(object):
    'Tree node.'
    def __init__(self,  taxid, name, level_rank, level_num, p_taxid, parent=None,children=None):
        self.taxid = taxid
        self.name = name
        self.level_rank= level_rank
        self.level_num = int(level_num)
        self.p_taxid = p_taxid
        self.all_reads = 0
        self.lvl_reads = 0
        #Parent/children attributes
        self.children = []
        self.parent = parent
        if children is not None:
            for child in children:
                self.add_child(child)
    def add_child(self, node):
        assert isinstance(node,Tree)
        self.children.append(node)
#################################################################################
#Main method
def main():
    #Parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('-i','--input', '-k','--kraken', dest='kraken_file', required=True,
        help='Kraken output file (5 tab-delimited columns, taxid in 3rd column)')
    parser.add_argument('-t','--taxonomy', dest='tax_file', required=True,
        help='Output taxonomy file from make_ktaxonomy.py')
    parser.add_argument('-o','--output',dest='out_file', required=True,
        help='Output kraken report file')
    parser.add_argument('--use-read-len',dest='use_read_len',
        action='store_true',default=False, required=False,
        help='Make report file using sum of read lengths [default: read counts]')
    args = parser.parse_args()

    #Start Program
    time = strftime("%m-%d-%Y %H:%M:%S", gmtime())
    sys.stdout.write("PROGRAM START TIME: " + time + '\n')

    #STEP 1/4: READ TAXONOMY FILE  
    count_nodes = 0
    sys.stdout.write(">> STEP 1/4: Reading taxonomy %s...\n" % args.tax_file)
    sys.stdout.write("\t%i nodes saved" % (count_nodes))
    #Parse taxonomy file 
    root_node = -1
    taxid2node = {}
    t_file = open(args.tax_file,'r')
    for line in t_file:
        count_nodes += 1
        sys.stdout.write("\r\t%i nodes saved" % (count_nodes))
        sys.stdout.flush()
        [taxid, p_tid, rank, lvl_num, name] = line.strip().split('\t|\t')
        curr_node = Tree(taxid, name, rank, lvl_num, p_tid)
        taxid2node[taxid] = curr_node
        #set parent/kids
        if taxid == "1":
            root_node = curr_node
        else:
            curr_node.parent = taxid2node[p_tid]
            taxid2node[p_tid].add_child(curr_node)
    t_file.close()
    sys.stdout.write("\r\t%i nodes saved\n" % (count_nodes))
    sys.stdout.flush()
    #STEP 2/4: READ KRAKEN FILE FOR COUNTS PER TAXID
    read_count = 0
    sys.stdout.write(">> STEP 2/4: Reading kraken file %s...\n" % args.kraken_file)
    sys.stdout.write("\t%i million reads processed" % read_count)
    sys.stdout.flush()
    #Save counts per taxid
    taxid2counts = {}
    taxid2allcounts = {} 
    k_file = open(args.kraken_file,'r')
    for line in k_file:
        read_count += 1
        if read_count % 1000 == 0:
            sys.stdout.write('\r\t%0.3f million reads processed' % float(read_count/1000000.))
            sys.stdout.flush()
        l_vals = line.strip().split('\t')
        taxid = l_vals[2]
        count = 1
        #If using read length instead of read counts
        if args.use_read_len:
            if '|' in l_vals[3]:
                [len1,len2] = l_vals[3].split('|')
                count = int(len1)+int(len2)
            else:
                count = int(l_vals[3])
        #add to dictionaries 
        if taxid not in taxid2counts:
            taxid2counts[taxid] = count
            taxid2allcounts[taxid] = count
        else:
            taxid2counts[taxid] += count
            taxid2allcounts[taxid] += count
    k_file.close()
    sys.stdout.write('\r\t%0.3f million reads processed\n' % float(read_count/1000000.))
    sys.stdout.flush()
    #STEP 3/4: FOR EVERY TAXID PARSED, ADD UP TOTAL READS
    sys.stdout.write(">> STEP 3/4: Creating final tree...\n")
    for curr_tid in taxid2counts:
        #Skip unclassified
        if curr_tid == '0':
            continue 
        p_node = taxid2node[curr_tid].parent 
        add_counts = taxid2counts[curr_tid] 
        #Assign reads for node
        taxid2node[curr_tid].lvl_reads += add_counts 
        taxid2node[curr_tid].all_reads += add_counts 
        while (p_node != None):
            #Add child reads to parent node 
            p_taxid = p_node.taxid
            if p_taxid not in taxid2allcounts:
                taxid2allcounts[p_taxid] = add_counts
            else:
                taxid2allcounts[p_taxid] += add_counts
            p_node.all_reads += add_counts
            #Get next parent node
            p_node = p_node.parent
    #STEP 4/4: PRINT REPORT FILE 
    sys.stdout.write(">> STEP 4/4: Printing report file to %s...\n" % args.out_file)
    o_file = open(args.out_file,'w')
    #Write line for unclassified reads:
    if '0'  in taxid2counts:
        o_file.write("%6.2f\t" % (float(taxid2counts['0'])/float(read_count)*100))
        o_file.write("%i\t%i\t" % (taxid2counts['0'],taxid2counts['0']))
        o_file.write('U\t0\tunclassified\n')
    #Get remaining lines 
    parse_nodes = [root_node]
    while len(parse_nodes) > 0:
        curr_node = parse_nodes.pop(0)
        curr_tid = curr_node.taxid 
        #Print information for this level
        o_file.write("%6.2f\t" % (float(taxid2allcounts[curr_tid])/float(read_count)*100))
        o_file.write("%i\t" % taxid2allcounts[curr_tid])
        if curr_tid not in taxid2counts:
            o_file.write("0\t")
        else:
            o_file.write("%i\t" % taxid2counts[curr_tid])
        o_file.write("%s\t" % curr_node.level_rank)
        o_file.write("%s\t" % curr_tid)
        o_file.write(" "*curr_node.level_num*2 + curr_node.name + "\n")
        #Add children to list
        for child in sorted(curr_node.children, key=operator.attrgetter('all_reads')):
            if child.taxid not in taxid2allcounts:
                continue
            if taxid2allcounts[child.taxid] == 0:
                continue
            #Add to list 
            parse_nodes.insert(0,child)    
    o_file.close() 
    #End of program
    time = strftime("%m-%d-%Y %H:%M:%S", gmtime())
    sys.stdout.write("PROGRAM END TIME: " + time + '\n')
    sys.exit(0)

#################################################################################
if __name__ == "__main__":
    main()
#################################################################################
##################################END OF PROGRAM#################################
#################################################################################




#!/usr/bin/env python
####################################################################
#combine_mpa.py converts multiple outputs from kreport2mpa.py
#Copyright (C) 2020 Jennifer Lu, jennifer.lu717@gmail.com

#This file is part of KrakenTools.
#KrakenTools is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 3 of the license, or
#(at your option) any later version.

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of 
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with this program; if not, see <http://www.gnu.org/licenses/>.

####################################################################
#Jennifer Lu, jlu26@jhmi.edu
#Updated: 07/12/2020
#
#This program reads multiple files in the 
#an mpa-format (MetaPhlAn) style report (as output from kreport2mpa.py).
#Each line represents a possible taxon classification. The first column is lists the 
#domain, kingdom, phyla, etc, leading up to each taxon.
#The levels are separated by the | delimiter, with the type of 
#level specified before each name with a single letter and underscore
#(d_ for domain, k_ for kingdom, etc). 
#The second column is the number of reads classified within 
#that taxon's subtree.
#
#Input files:
#   - MetaPhlAn format (mpa-format) files with two columns 
#   - All files must be generated from the same database, with the same
#     options from kreport2krona.py or errors may occur
#     
#Input Parameters to Specify [OPTIONAL]:
#   - header_line = prints a header line in mpa-report 
#       [Default: no header]
#   - intermediate-ranks = includes non-traditional taxon levels
#       (traditional levels: domain, kingdom, phylum, class, order, 
#       family, genus, species)
#       [Default: no intermediate ranks]
#Output file format (tab-delimited)
#   - Taxonomy tree levels |-delimited, with level type [d,k,p,c,o,f,g,s,x]
#   - Number of reads within subtree of the specified level
#
#Methods
#   - main
#
import os, sys, argparse

#Main method
def main():
    #Parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input', required=True,
        nargs='+', dest='in_files',
        help='Input files for this program (files generated by kreport2mpa.py)')
    parser.add_argument('-o', '--output', required=True,
        dest='o_file', help='Single mpa-report file name')
    args=parser.parse_args()

    #Process each file 
    samples = {} #Map number to name
    sample_count = 0 
    values = {} #Map taxon tree to sample to number 
    parent2child = {}
    toparse = []
    sys.stdout.write(" Number of files to parse: %i\n" % len(args.in_files))
    for in_file in args.in_files:
        i_file = open(in_file,'r')
        sample_count += 1
        sample_name = "Sample #" + str(sample_count) 
        for line in i_file:
            #Check for header line 
            if line[0] == "#":
                sample_name = line.strip().split('\t')[-1]
                continue 
            #Otherwise
            [classification, val] = line.strip().split('\t')
            #Check for parents
            split_vals = classification.split("|")
            curr_parent = ''
            for i in range(0,len(split_vals)):
                test_val = "|".join(split_vals[0:i]) 
                if test_val in values: 
                    curr_parent = test_val 
            #No parent
            if curr_parent == '':
                if classification not in values:
                    toparse.append(classification) 
            #Most specific parent found 
            if curr_parent != '':
                if curr_parent not in parent2child:
                    parent2child[curr_parent] = []
                if classification not in parent2child[curr_parent]:
                    parent2child[curr_parent].append(classification)
            #Save classification to value map
            if classification not in values:
                values[classification] = {}
            values[classification][sample_count] = val
        #Save sample name 
        samples[sample_count] = sample_name 
    
    sys.stdout.write(" Number of classifications to write: %i\n" % len(values))
    sys.stdout.write("\t%i classifications printed" % 0)
    #Write header
    o_file = open(args.o_file, 'w')
    o_file.write("#Classification") 
    for i in range(1, sample_count+1):
        o_file.write("\t" + samples[i])
    o_file.write("\n") 
    
    #Write each line  
    parsed = {} 
    count_c = 0 
    while len(toparse) > 0:
        curr_c = toparse.pop(0)
        #Add all children to stack 
        if curr_c in parent2child: 
            for child in parent2child[curr_c]: 
                toparse.insert(0, child) 
        #For the current classification, print per sample
        o_file.write(curr_c) 
        for i in range(1,sample_count + 1):
            if i in values[curr_c]:
                o_file.write("\t" + values[curr_c][i])
            else:
                o_file.write("\t0")
        o_file.write("\n")
        count_c += 1
        sys.stdout.write("\r\t%i classifications printed" % count_c)
        sys.stdout.flush()
    o_file.close() 
    sys.stdout.write("\r\t%i classifications printed\n" % count_c)
    sys.stdout.flush()

if __name__ == "__main__":
    main()



import os
import subprocess
import random
import argparse
from multiprocessing import Pool

def print_colorful_message(message, color):
    """
    Print a colorful message to the console.
    Args:
        message (str): The message to be printed.
        color (str): The color code to be applied.
                     Choices are 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white'.
    """
    colors = {
        'red': '\033[91m',
        'green': '\033[92m',
        'yellow': '\033[93m',
        'blue': '\033[94m',
        'magenta': '\033[95m',
        'cyan': '\033[96m',
        'white': '\033[97m',
    }
    end_color = '\033[0m'
    if color not in colors:
        print("Invalid color. Please choose from 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white'.")
        return
    colored_message = f"{colors[color]}{message}{end_color}"
    print(colored_message)

def process_sample(file, path1_fastq, path2_fastp, num_threads, suffix1, se, length_required):
    """
    Processes a single FASTQ file using fastp.
    Args:
        file (str): File name of the FASTQ file.
        path1_fastq (str): Path where raw FASTQ files are located.
        path2_fastp (str): Path where processed files will be saved.
        num_threads (int): Number of threads for fastp.
        suffix1 (str): Suffix for the forward reads files.
        se (bool): Flag to indicate if the data is single-end.
        length_required (int): Minimum length of reads to keep after processing.
    """
    suffix2 = suffix1.replace("1", "2")
    if file.endswith(suffix1):
        forward_file = os.path.join(path1_fastq, file)
        sample_id = file[:-len(suffix1)]
        output_forward = os.path.join(path2_fastp, file)
        task_file = os.path.join(path2_fastp, f"{sample_id}.task.complete")

        if os.path.exists(output_forward) and os.path.exists(task_file):
            print(f"Skipped: {forward_file}")
            return

        print_colorful_message(f">>> Processing: {sample_id}", "green")

        try:
            if se:
                command = [
                    "fastp", "-i", forward_file, "-o", output_forward,
                    "--thread", str(num_threads), "--length_required", str(length_required),
                    "--n_base_limit", "6", "--compression", "6",
                    "--html", f"{path2_fastp}/{sample_id}_fastp.html"
                ]
            else:
                reverse_file = forward_file[:-len(suffix1)] + suffix2
                output_reverse = output_forward[:-len(suffix1)] + suffix2
                command = [
                    "fastp", "-i", forward_file, "-o", output_forward,
                    "-I", reverse_file, "-O", output_reverse,
                    "--thread", str(num_threads), "--length_required", str(length_required),
                    "--n_base_limit", "6", "--compression", "6",
                    "--html", f"{path2_fastp}/{sample_id}_fastp.html"
                ]
            subprocess.run(command, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            with open(task_file, 'w') as f:
                f.write(">>> Processing complete for " + sample_id)
            print_colorful_message(f">>>> Task complete for {sample_id}", "cyan")
        except subprocess.CalledProcessError as e:
            print(f"Error processing {sample_id}: {e.stderr.decode()}")

def step1_fastq_qc(path1_fastq, path2_fastp, num_threads=8, suffix1="_1.fastq.gz", batch_size=5, se=False, length_required=50):
    """
    Preprocess FASTQ files using fastp in parallel.
    Args:
        path1_fastq (str): Path to raw FASTQ files.
        path2_fastp (str): Path to preprocessed FASTQ files.
        num_threads (int): Number of threads for fastp.
        suffix1 (str): Suffix for the forward reads files.
        batch_size (int): Number of samples to process simultaneously.
        se (bool): Flag to indicate if the sequencing data is single-end.
        length_required (int): Minimum length of reads to keep after processing.
    """
    print("   ")
    print_colorful_message("#########################################################", "blue")
    print_colorful_message(" ITMfinder: Identifing Intratumoral Microbiome pipeline ", "cyan")
    print_colorful_message(" If you encounter any issues, please report them at ", "cyan")
    print_colorful_message(" https://github.com/LiaoWJLab/ITMtools/issues ", "cyan")
    print_colorful_message("#########################################################", "blue")
    print(" Author: Dongqiang Zeng, Qianqian Mao ")
    print(" Email: interlaken@smu.edu.cn ")
    print_colorful_message("#########################################################", "blue")
    print("   ")
    
    print(" >>> Step-1 FASTQ files quality control using fastp...  ")

    os.makedirs(path2_fastp, exist_ok=True)
    fastq_files = [f for f in os.listdir(path1_fastq) if f.endswith(suffix1)]
    random.shuffle(fastq_files)

    with Pool(processes=batch_size) as pool:
        pool.starmap(process_sample, [(file, path1_fastq, path2_fastp, num_threads, suffix1, se, length_required) for file in fastq_files])


# Keep the existing code as it is, but add a main function for import purposes
def main():
    parser = argparse.ArgumentParser(description="Preprocess FASTQ files using fastp")
    parser.add_argument("--path1_fastq", type=str, required=True, help="Path to raw FASTQ files")
    parser.add_argument("--path2_fastp", type=str, required=True, help="Path to preprocessed FASTQ files")
    parser.add_argument("--num_threads", type=int, default=8, help="Number of threads for fastp")
    parser.add_argument("--suffix1", type=str, default="_1.fastq.gz", help="Suffix of the forward reads file")
    parser.add_argument("--batch_size", type=int, default=5, help="Number of samples to process simultaneously")
    parser.add_argument("--se", action='store_true', help="Indicate if the sequencing data is single-end")
    parser.add_argument("--length_required", type=int, default=50, help="Minimum length of reads to keep after processing")
    args = parser.parse_args()

    step1_fastq_qc(args.path1_fastq, args.path2_fastp, args.num_threads, args.suffix1, args.batch_size, args.se, args.length_required)

# Ensure the main function only runs if this file is executed directly
if __name__ == "__main__":
    main()



# file dir：itmfinder/itm_helper/__init__.py
from .alpha_diversity import main as alpha_diversity_main
from .beta_diversity import main as beta_diversity_main
from .combine_bracken_outputs import main as combine_bracken_main
from .combine_kreports import main as combine_kreports_main
from .combine_mpa import main as combine_mpa_main
from .extract_kraken_reads import extract_reads_i
from .kraken2otu import main as kraken2otu_main
from .kreport2krona import main as kreport2krona_main
from .kreport2mpa import kreport2mpai
from .make_ktaxonomy import main as make_ktaxonomy_main
from .make_kreport import main as make_kreport_main



import os
import subprocess
import argparse
import random
from multiprocessing import Pool
from itmfinder.itm_helper.kreport2mpa import kreport2mpai
from itmfinder.itm_helper.alpha_diversity import main as alpha_diversity_main

def print_colorful_message(message, color):
    """
    Print a colorful message to the console.
    Args:
        message (str): The message to be printed.
        color (str): The color code to be applied.
    """
    colors = {
        'red': '\033[91m',
        'green': '\033[92m',
        'yellow': '\033[93m',
        'blue': '\033[94m',
        'magenta': '\033[95m',
        'cyan': '\033[96m',
        'white': '\033[97m',
    }
    end_color = '\033[0m'
    print(colors.get(color, '\033[97m') + message + end_color)

def process_sample(sample_id, paths, db_ku, num_threads, force):
    path6_rcr, path7_ku2, path8_bracken = paths
    kraken_output = os.path.join(path7_ku2, f"{sample_id}.kraken.output.txt")
    kraken_report = os.path.join(path7_ku2, f"{sample_id}.kraken.report.txt")
    task_complete_file = os.path.join(path7_ku2, f"{sample_id}.task.complete")

    # Check if processing is already completed
    if os.path.exists(task_complete_file) and os.path.exists(kraken_report) and not force:
        print_colorful_message(f"Sample {sample_id} processing already completed. Skipping...", "green")
        return

    print_colorful_message(f">>> Running Kraken2 and Bracken for sample: {sample_id}", "yellow")

    # Paths for paired-end fastq files
    fastq_files = [os.path.join(path6_rcr, f"{sample_id}_rcr_1.fastq.gz"), os.path.join(path6_rcr, f"{sample_id}_rcr_2.fastq.gz")]

    # Run Kraken2 for taxonomic classification
    subprocess.run(["kraken2", "--db", db_ku, "--threads", str(num_threads), "--paired", "--report-minimizer-data",
                    "--report", kraken_report, "--output", kraken_output] + fastq_files)

    # Convert Kraken report to MPA format
    kreport2mpai(report_path=kraken_report, output_path=os.path.join(path7_ku2, f"{sample_id}.kraken.mpa.txt"))

    # Define paths for Bracken outputs at various taxonomic levels
    bracken_outputs = {
        "G": os.path.join(path8_bracken, f"{sample_id}.g.bracken"),
        "S": os.path.join(path8_bracken, f"{sample_id}.s.bracken"),
        "F": os.path.join(path8_bracken, f"{sample_id}.f.bracken"),
        "O": os.path.join(path8_bracken, f"{sample_id}.o.bracken")
    }

    # Run Bracken for abundance estimation at different taxonomic levels
    for level, output in bracken_outputs.items():
        subprocess.run(["bracken", "-d", db_ku, "-i", kraken_report, "-o", output, "-r", "100", "-l", level, "-t", "2"])

    # Calculate alpha diversity for genus and species levels
    diversity_outputs = {
        "G": os.path.join(path8_bracken, f"{sample_id}.diversity.g.txt"),
        "S": os.path.join(path8_bracken, f"{sample_id}.diversity.s.txt")
    }

    # Execute alpha diversity calculations for each level
    for level, bracken_output in bracken_outputs.items():
        if level in diversity_outputs:
            diversity_output = diversity_outputs[level]
            with open(diversity_output, "a") as f:
                for metric in ["Sh", "BP", "Si", "ISi", "F"]:
                    # Call alpha_diversity_main with expected parameters
                    alpha_diversity_main(bracken_output, metric)

    # Mark task as completed
    with open(task_complete_file, "w") as f:
        f.write("Processing completed.")

def step6_kraken2Bracken(path6_rcr, path7_ku2, path8_bracken, db_ku, num_threads=8, force=False):
    print("   ")
    print_colorful_message("#########################################################", "blue")
    print_colorful_message(" ITMfinder: Identifying Intratumoral Microbiome pipeline ", "cyan")
    print_colorful_message(" If you encounter any issues, please report them at ", "cyan")
    print_colorful_message(" https://github.com/LiaoWJLab/ITMfinder/issues ", "cyan")
    print_colorful_message("#########################################################", "blue")
    print(" Author: Dongqiang Zeng, Qianqian Mao ")
    print(" Email: interlaken@smu.edu.cn ")
    print_colorful_message("#########################################################", "blue")
    print("   ")

    print(" >>> Step-6 Perform taxonomic classification using Kraken2...  ")

    os.makedirs(path7_ku2, exist_ok=True)
    os.makedirs(path8_bracken, exist_ok=True)

    # List all samples
    mr_files = [f for f in os.listdir(path6_rcr) if f.endswith("_rcr_1.fastq.gz")]
    random.shuffle(mr_files)  # Shuffle for random processing order
    paths = (path6_rcr, path7_ku2, path8_bracken)
    
    # Process samples in parallel
    with Pool() as pool:
        pool.starmap(process_sample, [(f.split('_')[0], paths, db_ku, num_threads, force) for f in mr_files])

    print_colorful_message(">>> All samples have been processed.", "blue")

def main():
    parser = argparse.ArgumentParser(description="Step 6: Taxonomic classification using Kraken2 for report")
    parser.add_argument("--path6_rcr", type=str, required=True, help="Path to microbiome reads after decontamination")
    parser.add_argument("--path7_ku2", type=str, required=True, help="Path to store Kraken2 outputs")
    parser.add_argument("--path8_bracken", type=str, required=True, help="Path for Bracken outputs")
    parser.add_argument("--db_ku", type=str, required=True, help="Path to Kraken2 database")
    parser.add_argument("--num_threads", type=int, default=8, help="Number of threads")
    parser.add_argument("--force", action="store_true", help="Force execution of downstream Kraken2 process")
    args = parser.parse_args()

    step6_kraken2Bracken(args.path6_rcr, args.path7_ku2, args.path8_bracken, args.db_ku, args.num_threads, args.force)

# Ensure the main function only runs if this file is executed directly
if __name__ == "__main__":
    main()



# File path: itmfinder/utils/data_loader.py
import importlib.resources as pkg_resources

def load_contamination_list():
    """
    Load the contamination list from the package resources.
    This function reads the 'conta_list.txt' from the 'resources' directory,
    parsing it to extract a list of taxonomic IDs.
    
    Returns:
        list of str: A list of taxonomic IDs.
    """
    try:
        # Open the resource from the package using importlib.resources
        with pkg_resources.files('itmfinder.resources').joinpath('conta_list.txt').open('r') as file:
            lines = file.readlines()
        
        # Extract the second column (taxonomic IDs) from each line if it's numeric
        tax_ids = [line.strip().split()[1] for line in lines if len(line.strip().split()) > 1 and line.strip().split()[1].isdigit()]
        return tax_ids
    except FileNotFoundError:
        print("The contamination list file 'conta_list.txt' was not found in 'resources'.")
        return []
    except IndexError:
        print("Failed to parse taxonomic IDs from 'conta_list.txt'. Please ensure the file format is correct.")
        return []
    except Exception as e:
        print(f"An unexpected error occurred while loading the contamination list: {e}")
        return []



#!/usr/bin/env python
################################################################
#combine_kreports.py takes multiple kraken-style reports and combines
#them into a single report file
#Copyright (C) 2019-2020 Jennifer Lu, jennifer.lu717@gmail.com
#
#This file is part of KrakenTools
#KrakenTools is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 3 of the license, or 
#(at your option) any later version.

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of 
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program; if not, see <http://www.gnu.org/licenses/>.

#################################################################
#Jennifer Lu, jlu26@jhmi.edu
#Updated: 05/16/2019
#
#This program reads in multiple Kraken report files and generates
#a combined Kraken report with columns for read counts and summarized
#read counts for each sample, along with two columns for across-sample sums
#
#Parameters:
#   -h, --help................show help message.
#   -r X, --report-file X.....all input kraken reports (separated by spaces)
#   -o X, --output X..........output kraken report filename
#   --display-headers.........includes header lines mapping samples to abbreviated names
#                             [default:true]
#   --no-headers..............do not include header lines [default:false] 
#   --sample-names............sample names for each kraken report (separated by spaces)
#                             [if none are given, each sample is given names S1, S2, etc] 
#Each Input report file format (tab-delimited)
#   - percentage of total reads
#   - number of reads (including reads within subtree)
#   - number of reads (only at this level)
#   - taxonomic classification level (U, D, P, C, O, F, G, S,...etc)
#   - NCBI taxonomic ID
#   - name of level
#Output file format (tab-delimited)
#   - percentage of total reads (for summed reads)
#   - combined number of reads (including reads within subtree)
#   - combined number of reads (only at this level)
#   - S1_all_reads, S1_lvl_reads, S2_all_reads, S2_lvl_reads, ...etc.
#   - taxonomic classification level (U, D, P, C, O, F, G, S,...etc)
#   - NCBI taxonomic ID
#   - name of level
#Methods 
#   - main
#   - process_kraken_report
####################################################################
import os, sys, argparse
import operator
from time import gmtime 
from time import strftime 

#Tree Class 
#usage: tree node used in constructing a taxonomy tree
#   including only the taxonomy levels and genomes identified in the Kraken report
class Tree(object):
    'Tree node.'
    def __init__(self, name, taxid, level_num, level_id, all_reads, lvl_reads, children=None, parent=None):
        self.name = name
        self.taxid = taxid
        self.level_num = level_num
        self.level_id = level_id
        self.tot_all = all_reads
        self.tot_lvl = lvl_reads
        self.all_reads = {}
        self.lvl_reads = {}
        self.children = []
        self.parent = parent
        if children is not None:
            for child in children:
                self.add_child(child)
    def add_child(self,node):
        assert isinstance(node,Tree)
        self.children.append(node)
    def add_reads(self, sample, all_reads, lvl_reads):
        self.all_reads[sample] = all_reads
        self.lvl_reads[sample] = lvl_reads
        self.tot_all += all_reads
        self.tot_lvl += lvl_reads
    def __lt__(self,other):
        return self.tot_all < other.tot_all
         
####################################################################
#process_kraken_report
#usage: parses a single line in the kraken report and extracts relevant information
#input: kraken report file with the following tab delimited lines 
#   - percent of total reads   
#   - number of reads (including at lower levels)
#   - number of reads (only at this level)
#   - taxonomy classification of level 
#       (U, - (root), - (cellular org), D, P, C, O, F, G, S) 
#   - taxonomy ID (0 = unclassified, 1 = root, 2 = Bacteria...etc)
#   - spaces + name 
#returns:
#   - classification/genome name
#   - taxonomy ID for this classification
#   - level for this classification (number)
#   - level name (U, -, D, P, C, O, F, G, S)
#   - all reads classified at this level and below in the tree
#   - reads classified only at this level
def process_kraken_report(curr_str):
    split_str = curr_str.strip().split('\t')
    if len(split_str) < 5:
        return []
    try:
        int(split_str[1])
    except ValueError:
        return []
    #Extract relevant information
    all_reads =  int(split_str[1])
    level_reads = int(split_str[2])
    try:
        level_type = split_str[-3]
        taxid = int(split_str[-2])
    except ValueError:
        taxid = int(split_str[-3])
        level_type = split_str[-2]
        map_kuniq = {'species':'S','genus':'G','family':'F',
            'order':'O','class':'C','phylum':'P','superkingdom':'D',
            'kingdom':'K'}
        if level_type not in map_kuniq:
            level_type = '-'
        else:
            level_type= map_kuniq[level_type]
    #Get name and spaces
    spaces = 0
    name = split_str[-1]
    for char in name:
        if char == ' ':
            name = name[1:]
            spaces += 1 
        else:
            break 
    #Determine which level based on number of spaces
    level_num = int(spaces/2)
    return [name, taxid, level_num, level_type, all_reads, level_reads]
    
####################################################################
#Main method
def main():
    #Parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('-r','--report-file','--report-files',
        '--report','--reports', required=True,dest='r_files',nargs='+',
        help='Input kraken report files to combine (separate by spaces)') 
    parser.add_argument('-o','--output', required=True,dest='output',
        help='Output kraken report file with combined information')
    parser.add_argument('--display-headers',required=False,dest='headers',
        action='store_true', default=True,
        help='Include header lines')
    parser.add_argument('--no-headers',required=False,dest='headers',
        action='store_false',default=True,
        help='Do not include header lines')
    parser.add_argument('--sample-names',required=False,nargs='+',
        dest='s_names',default=[],help='Sample names to use as headers in the new report')
    parser.add_argument('--only-combined', required=False, dest='c_only',
        action='store_true', default=False, 
        help='Include only the total combined reads column, not the individual sample cols')
    args=parser.parse_args()
    

    #Initialize combined values 
    main_lvls = ['U','R','D','K','P','C','O','F','G','S']
    map_lvls = {'kingdom':'K', 'superkingdom':'D','phylum':'P','class':'C','order':'O','family':'F','genus':'G','species':'S'}
    count_samples = 0
    num_samples = len(args.r_files)
    sample_names = args.s_names
    root_node = -1 
    prev_node = -1
    curr_node = -1
    u_reads = {0:0} 
    total_reads = {0:0} 
    taxid2node = {}

    #Check input values 
    if len(sample_names) > 0 and len(sample_names) != num_samples: 
        sys.stderr.write("Number of sample names provided does not match number of reports\n")
        sys.exit(1)
    #Map names
    id2names = {} 
    id2files = {} 
    if len(sample_names) == 0:
        for i in range(num_samples):
            id2names[i+1] = "S" + str(i+1)
            id2files[i+1] = ""
    else:
        for i in range(num_samples):
            id2names[i+1] = sample_names[i] 
            id2files[i+1] = ""
    
    #################################################
    #STEP 1: READ IN REPORTS
    #Iterate through reports and make combined tree! 
    sys.stdout.write(">>STEP 1: READING REPORTS\n")
    sys.stdout.write("\t%i/%i samples processed" % (count_samples, num_samples))
    sys.stdout.flush()
    for r_file in args.r_files:
        count_samples += 1 
        sys.stdout.write("\r\t%i/%i samples processed" % (count_samples, num_samples))
        sys.stdout.flush()
        id2files[count_samples] = r_file
        #Open File 
        curr_file = open(r_file,'r')
        for line in curr_file: 
            report_vals = process_kraken_report(line)
            if len(report_vals) < 5:
                continue
            [name, taxid, level_num, level_id, all_reads, level_reads] = report_vals
            if level_id in map_lvls:
                level_id = map_lvls[level_id]
            #Total reads 
            total_reads[0] += level_reads
            total_reads[count_samples] = level_reads 
            #Unclassified 
            if level_id == 'U' or taxid == '0':
                u_reads[0] += level_reads
                u_reads[count_samples] = level_reads 
                continue
            #Tree Root 
            if taxid == 1: 
                if count_samples == 1:
                    root_node = Tree(name, taxid, level_num, 'R', 0,0)
                    taxid2node[taxid] = root_node 
                root_node.add_reads(count_samples, all_reads, level_reads) 
                prev_node = root_node
                continue 
            #Move to correct parent
            while level_num != (prev_node.level_num + 1):
                prev_node = prev_node.parent
            #IF NODE EXISTS 
            if taxid in taxid2node: 
                taxid2node[taxid].add_reads(count_samples, all_reads, level_reads) 
                prev_node = taxid2node[taxid]
                continue 
            #OTHERWISE
            #Determine correct level ID
            if level_id == '-' or len(level_id)> 1:
                if prev_node.level_id in main_lvls:
                    level_id = prev_node.level_id + '1'
                else:
                    num = int(prev_node.level_id[-1]) + 1
                    level_id = prev_node.level_id[:-1] + str(num)
            #Add node to tree
            curr_node = Tree(name, taxid, level_num, level_id, 0, 0, None, prev_node)
            curr_node.add_reads(count_samples, all_reads, level_reads)
            taxid2node[taxid] = curr_node
            prev_node.add_child(curr_node)
            prev_node = curr_node 
        curr_file.close()

    sys.stdout.write("\r\t%i/%i samples processed\n" % (count_samples, num_samples))
    sys.stdout.flush()

    #################################################
    #STEP 2: SETUP OUTPUT FILE
    sys.stdout.write(">>STEP 2: WRITING NEW REPORT HEADERS\n")
    o_file = open(args.output,'w') 
    #Lines mapping sample ids to filenames
    if args.headers: 
        o_file.write("#Number of Samples: %i\n" % num_samples) 
        o_file.write("#Total Number of Reads: %i\n" % total_reads[0])
        for i in id2names:
            o_file.write("#")
            o_file.write("%s\t" % id2names[i])
            o_file.write("%s\n" % id2files[i])
        #Report columns
        o_file.write("#perc\ttot_all\ttot_lvl")
        if not args.c_only:
            for i in id2names:
                o_file.write("\t%s_all" % id2names[i])
                o_file.write("\t%s_lvl" % id2names[i])
        o_file.write("\tlvl_type\ttaxid\tname\n")
    #################################################
    #STEP 3: PRINT TREE
    sys.stdout.write(">>STEP 3: PRINTING REPORT\n")
    #Print line for unclassified reads
    o_file.write("%0.4f\t" % (float(u_reads[0])/float(total_reads[0])*100))
    for i in range(num_samples+1):
        if i == 0 or (i > 0 and not args.c_only):
            if i not in u_reads:
                o_file.write("0\t")
                o_file.write("0\t")
            else:
                o_file.write("%i\t" % u_reads[i])
                o_file.write("%i\t" % u_reads[i])
    o_file.write("U\t0\tunclassified\n")
    #Print for all remaining reads 
    all_nodes = [root_node]
    curr_node = -1
    curr_lvl = 0
    prev_node = -1
    while len(all_nodes) > 0:
        #Remove node and insert children
        curr_node = all_nodes.pop()
        if len(curr_node.children) > 0:
            curr_node.children.sort()
            for node in curr_node.children:
                all_nodes.append(node)
        #Print information for this node 
        o_file.write("%0.4f\t" % (float(curr_node.tot_all)/float(total_reads[0])*100))
        o_file.write("%i\t" % curr_node.tot_all)
        o_file.write("%i\t" % curr_node.tot_lvl)
        if not args.c_only:
            for i in range(num_samples):
                if (i+1) not in curr_node.all_reads: 
                    o_file.write("0\t0\t")
                else:
                    o_file.write("%i\t" % curr_node.all_reads[i+1])
                    o_file.write("%i\t" % curr_node.lvl_reads[i+1])
        o_file.write("%s\t" % curr_node.level_id)
        o_file.write("%s\t" % curr_node.taxid)
        o_file.write(" "*curr_node.level_num*2)
        o_file.write("%s\n" % curr_node.name)
    o_file.close() 
####################################################################
if __name__ == "__main__":
    main()



import os
import subprocess
import random
import argparse
from multiprocessing import Pool
from functools import partial

def print_colorful_message(message, color):
    """
    Print a colorful message to the console.

    Args:
        message (str): The message to be printed.
        color (str): The color code to be applied. 
                     'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white'
    """
    colors = {
        'red': '\033[91m',
        'green': '\033[92m',
        'yellow': '\033[93m',
        'blue': '\033[94m',
        'magenta': '\033[95m',
        'cyan': '\033[96m',
        'white': '\033[97m',
    }
    end_color = '\033[0m'
    if color not in colors:
        print("Invalid color. Please choose from 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white'.")
        return
    colored_message = f"{colors[color]}{message}{end_color}"
    print(colored_message)

def process_sample(file, path3_hr, path3_hr2, db_bowtie2, suffix1, num_threads, se):
    """
    Process a single sample to remove host reads using Bowtie2.
    """
    try:
        # Extract sample ID from the filename
        sample_id = os.path.basename(file).replace(suffix1, "")  # Remove the suffix to get sample ID
        output = f"{sample_id}_host_remove"
        log_file = os.path.join(path3_hr2, f"{sample_id}_bowtie2.log") 

        # FOR SINGLE-END
        if se:
            output_file = f"{output}.fastq.gz"
            # Check if the output file already exists
            if os.path.exists(os.path.join(path3_hr2, output_file)):
                print(f"Skipped: {output_file} as it already exists.")
                return
            
            print(f" >>> Processing file: {file}")
            
            # Run Bowtie2 to remove host reads
            subprocess.run(["bowtie2", "-p", str(num_threads), "-x", db_bowtie2, "-U", os.path.join(path3_hr, file), 
                           "--un-gz", os.path.join(path3_hr2, output), "--no-unal"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

            # Rename the output file
            os.rename(os.path.join(path3_hr2, f"{output}.fastq.gz"), os.path.join(path3_hr2, output_file))
            
        # FOR PAIRED-END
        else:
            output_file1 = f"{output}_R1.fastq.gz"
            output_file2 = f"{output}_R2.fastq.gz"

            # Check if the output files already exist
            if os.path.exists(os.path.join(path3_hr2, output_file1)) and os.path.exists(os.path.join(path3_hr2, output_file2)):
                print(f"Skipped: {output_file1} and {output_file2} as they already exist.")
                return
            
            # Determine the second-end filename
            sample_id_second_end = f"{output_file2}"

            print(f" >>> Processing files: {file} and {sample_id_second_end}")

            subprocess.run(["bowtie2", "-p", str(num_threads), "-x", db_bowtie2, "-1", os.path.join(path3_hr, file), 
                             "-2", os.path.join(path3_hr, sample_id_second_end), "--un-conc-gz", os.path.join(path3_hr2, output), "--no-unal"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

            # Rename the output files
            os.rename(os.path.join(path3_hr2, f"{output}.1"), os.path.join(path3_hr2, output_file1))
            os.rename(os.path.join(path3_hr2, f"{output}.2"), os.path.join(path3_hr2, output_file2))
        
        # Create task.complete file
        open(os.path.join(path3_hr2, f"{sample_id}.task.complete"), 'a').close()

    except Exception as e:
        print(f"Error processing file {file}: {e}")

def step2_remove_host_reads2(path3_hr, path3_hr2, db_bowtie2, suffix1="_host_remove_R1.fastq.gz", num_threads=8, batch_size=1, se=False):
    """
    Remove host reads from preprocessed FASTQ files using Bowtie2.

    Args:
        path3_hr (str): Path to preprocessed FASTQ files.
        path3_hr2 (str): Path to host removed FASTQ files.
        db_bowtie2 (str): Path to Bowtie2 database.
        suffix1 (str): Suffix of the forward reads file. Default is "_host_remove_R1.fastq.gz".
        num_threads (int): Number of threads to use for Bowtie2 alignment. Default is 8.
        batch_size (int): Batch size for concurrent processing. Default is 1.
        se (bool): Whether the data is single-end sequencing. Default is False.
    """

    print("   ")
    print_colorful_message("#########################################################", "blue")
    print_colorful_message(" ITMfinder: Identifing Intratumoral Microbiome pipeline ", "cyan")
    print_colorful_message(" If you encounter any issues, please report them at ", "cyan")
    print_colorful_message(" https://github.com/LiaoWJLab/ITMfinder/issues ", "cyan")
    print_colorful_message("#########################################################", "blue")
    print(" Author: Dongqiang Zeng, Qianqian Mao ")
    print(" Email: interlaken@smu.edu.cn ")
    print_colorful_message("#########################################################", "blue")
    print("   ")

    print(" >>> Step-2 Remove host gene reads secondly...  ")

    # Create the output directory if it does not exist
    os.makedirs(path3_hr2, exist_ok=True)
    
    # Get the list of files in the input directory
    files = sorted([file for file in os.listdir(path3_hr) if file.endswith(suffix1)], reverse=True)
    random.shuffle(files)  # Shuffle the files randomly
    total_files = len(files)
    
    # Define partial function with fixed arguments
    partial_process_sample = partial(process_sample, path3_hr=path3_hr, path3_hr2=path3_hr2, db_bowtie2=db_bowtie2, suffix1=suffix1, num_threads=num_threads, se=se)
    
    # Iterate through the files in the input directory
    for index in range(0, total_files, batch_size):
        batch_files = files[index:index+batch_size]
        with Pool(processes=batch_size) as pool:
            pool.map(partial_process_sample, batch_files)

def main():
    parser = argparse.ArgumentParser(description="Step 2: Remove host reads using Bowtie2")
    parser.add_argument("--path3_hr", type=str, help="Path to host removed FASTQ files firstly")
    parser.add_argument("--path3_hr2", type=str, help="Path to host removed FASTQ files secondly")
    parser.add_argument("--db_bowtie2", type=str, help="Path to Bowtie2 database")
    parser.add_argument("--num_threads", type=int, default=8, help="Number of threads")
    parser.add_argument("--suffix1", type=str, default="_host_remove_R1.fastq.gz", help="Suffix of the forward reads file")
    parser.add_argument("--batch_size", type=int, default=1, help="Batch size for concurrent processing")
    parser.add_argument("--se", action='store_true', help="Single-end sequencing data")
    args = parser.parse_args()

    step2_remove_host_reads2(args.path3_hr, args.path3_hr2, args.db_bowtie2, args.suffix1, args.num_threads, args.batch_size, args.se)

# Ensure the main function only runs if this file is executed directly
if __name__ == "__main__":
    main()




import os
import pandas as pd
import argparse

def print_colorful_message(message, color):
    """
    Print a colorful message to the console.

    Args:
        message (str): The message to be printed.
        color (str): The color code to be applied. 
                     'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white'
    """
    colors = {
        'red': '\033[91m',
        'green': '\033[92m',
        'yellow': '\033[93m',
        'blue': '\033[94m',
        'magenta': '\033[95m',
        'cyan': '\033[96m',
        'white': '\033[97m',
    }
    end_color = '\033[0m'
    if color not in colors:
        print("Invalid color. Please choose from 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white'.")
        return
    colored_message = f"{colors[color]}{message}{end_color}"
    print(colored_message)


def get_file_size(filepath):
    """Get file size in MB"""
    if not os.path.exists(filepath):  # 检查文件是否存在
        raise FileNotFoundError(f"File '{filepath}' does not exist.")
    
    if not os.path.isfile(filepath):  # 检查路径是否指向一个文件
        raise ValueError(f"'{filepath}' is not a valid file path.")
    
    # 获取文件大小并返回
    return round(os.path.getsize(filepath) / (1024 * 1024), 2)
    

def get_files_and_sizes(directory, suffix):
    """Get filenames and sizes for files with specified suffix in directory"""
    files_and_sizes = []
    for file in os.listdir(directory):
        if file.endswith(suffix):
            filepath = os.path.join(directory, file)
            file_size = get_file_size(filepath)
            files_and_sizes.append((file, file_size))
    return files_and_sizes

def process_data_for_sample(sample_id, path2_fastp, path3_hr, path7_ku2, path6_rcr):
    """Process files for a single sample"""
    sample_data = []
    
    # Process files for path2_fastp
    fastp_files = get_files_and_sizes(os.path.join(path2_fastp), f"{sample_id}_1.fastq.gz")  # Only process single-end files
    # print("Fastp files:", fastp_files)
    
    # Process files for path3_hr
    hr_files = get_files_and_sizes(os.path.join(path3_hr), f"{sample_id}_host_remove_R1.fastq.gz")  # Only process single-end files
    # print("Host-removed files:", hr_files)
    
    # Process files for path6_rcr
    mr_files = get_files_and_sizes(os.path.join(path6_rcr), f"{sample_id}_rcr_1.fastq")  # Only process single-end files
    # print("Microbiome files:", mr_files)

    # Process files for path7_ku2
    ku2_file = get_files_and_sizes(os.path.join(path7_ku2), f"{sample_id}.kraken.output.txt")
    # print("Kraken files:", ku2_file)

    # Ensure all lists have the same length
    min_length = min(len(fastp_files), len(hr_files), len(mr_files))
    
    # Combine fastp_files, hr_files, and mr_files using zip and iterate over them
    for i in range(min_length):
        sample_data.append((sample_id, fastp_files[i][0], fastp_files[i][1], hr_files[i][0], hr_files[i][1], ku2_file[0][0], ku2_file[0][1], mr_files[i][0], mr_files[i][1]))
    
    return sample_data

def step8_get_data_summary(path2_fastp, path3_hr, path7_ku2, path6_rcr, out_dir):
    """Process files for each sample and organize into a dataframe"""

    print("   ")
    print_colorful_message("#########################################################", "blue")
    print_colorful_message(" ITMfinder: Identifing Intratumoral Microbiome pipline", "cyan")
    print_colorful_message(" If you encounter any issues, please report them at ", "cyan")
    print_colorful_message(" https://github.com/LiaoWJLab/ITMtools/issues ", "cyan")
    print_colorful_message("#########################################################", "blue")
    print(" Author: Dongqiang Zeng, Qianqian Mao ")
    print(" Email: interlaken@smu.edu.cn ")
    print("   ")

    print(" >>> Step-8 Generate Data Summary...  ")

    all_sample_data = []
    for file in os.listdir(path2_fastp):
        # Extract sample_id by removing suffixes like '_1.fastq.gz'
        sample_id = file.replace("_1.fastq.gz", "").replace("_2.fastq.gz", "")
        
        # Process data for the current sample
        sample_data = process_data_for_sample(sample_id, path2_fastp, path3_hr, path7_ku2, path6_rcr)
        all_sample_data.extend(sample_data)

    # Create a dataframe
    df = pd.DataFrame(all_sample_data, columns=["Sample", "Fastp_File", "Fastp_Size_MB",
                                                 "Host_Removed_File", "HR_Size_MB",
                                                 "Kraken_Output_File", "KU2_Size_MB",
                                                 "Microbiome_files", "Microbiome_Size_MB"])
    
    # Remove duplicate rows based on the first column 'Sample'
    df.drop_duplicates(subset=['Sample'], inplace=True)
    
    # Sort dataframe by 'Microbiome_Size_MB' column in descending order
    df.sort_values(by='Microbiome_Size_MB', ascending=False, inplace=True)
    
    print("Dataframe:")
    print(df)
    
    # Save the dataframe to the output directory
    output_file = os.path.join(out_dir, "0-sample_file_sizes.csv")
    df.to_csv(output_file, index=False)
    print(f"Sample file sizes saved to: {output_file}")

def main():
    parser = argparse.ArgumentParser(description="Process files for each sample and organize into a dataframe")
    parser.add_argument("--path2_fastp", type=str, help="Path to preprocessed FASTQ files")
    parser.add_argument("--path3_hr", type=str, help="Path to host-removed FASTQ files")
    parser.add_argument("--path7_ku2", type=str, help="Path to Kraken output files")
    parser.add_argument("--path6_rcr", type=str, help="Path to microbiome reads")
    parser.add_argument("--out_dir", type=str, help="Output directory")
    args = parser.parse_args()

    step8_get_data_summary(args.path2_fastp, args.path3_hr, args.path7_ku2, args.path6_rcr, args.out_dir)

# Ensure the main function only runs if this file is executed directly
if __name__ == "__main__":
    main()




#!/usr/bin/env python
######################################################################
#make_taxonomy.py takes in three files: nodes.dmp, names.dmp, and 
#seqid2taxid.map to create a condensed taxonomy file 
#Copyright (C) 2020 Jennifer Lu, jennifer.lu717@gmail.com
#
#This file is part of KrakenTools
#KrakenTools is free software; oyu can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 3 of the license, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of 
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program; if not, see <http://www.gnu.org/licenses/>.
#
######################################################################
#Jennifer Lu, jlu26@jhmi.edu
#Updated: 04/14/2020
#
#This program creates a condensed taxonomy file for a given Kraken database
#To create a taxonomy, the nodes.dmp, names.dmp, and seqid2taxid.map files
#must be provided
#
#Required Parameters:
#   --nodes X...........................nodes.dmp file
#   --names X...........................names.dmp file 
#   --seqid2taxid X.....................seqid2taxid.map file
#   -o, --output X......................output file with taxonomy info
#Optional Parameters:
#   -h, --help..........................show help message.
#################################################################################
import os, sys, argparse
from time import gmtime
from time import strftime 
#################################################################################
#Tree Class
#usage: tree node used in constructing taxonomy tree
class Tree(object):
    'Tree node.'
    def __init__(self,  taxid, level_rank, parent=None,children=None):
        self.taxid = taxid
        self.level_rank= level_rank
        #Other attributes for later
        self.name = ''
        self.level_num = -1
        self.p_taxid = -1
        #Parent/children attributes
        self.children = []
        self.parent = parent
        if children is not None:
            for child in children:
                self.add_child(child)
    def add_child(self, node):
        assert isinstance(node,Tree)
        self.children.append(node)
#################################################################################
#Main method
def main():
    #Parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('--nodes',dest='nodes_file', required=True,
        help='nodes.dmp file from taxonomy')
    parser.add_argument('--names',dest='names_file', required=True,
        help='names.dmp file from taxonomy')
    parser.add_argument('--seqid2taxid',dest='s2t_file', required=True,
        help='seqid2taxid.map file')
    parser.add_argument('-o','--output',dest='out_file', required=True,
        help='output taxonomy file')
    args = parser.parse_args()

    #Start Program
    time = strftime("%m-%d-%Y %H:%M:%S", gmtime())
    sys.stdout.write("PROGRAM START TIME: " + time + '\n')

    map_ranks = {'superkingdom':'D',
        'phylum':'P',
        'class':'C',
        'order':'O',
        'family':'F',
        'genus':'G',
        'species':'S'} 
    #STEP 1/5: PARSE NODES.DMP FILE
    root_node = -1
    taxid2node = {} 
    p_notsaved = {} 
    nodes_f = open(args.nodes_file,'r')
    sys.stdout.write(">> STEP 1/5: Reading %s\n" % args.nodes_file)
    sys.stdout.write("\t%0 nodes read")
    count_nodes = 0 
    for line in nodes_f:
        count_nodes += 1
        if (count_nodes % 100) == 0:
            sys.stdout.write("\r\t%i nodes read" % count_nodes)
            sys.stdout.flush()
        [curr_taxid,parent_taxid,rank] = line.strip().split("\t|\t")[0:3]
        #Make/Save node
        newrank = '-'
        if rank in map_ranks:
            newrank = map_ranks[rank]
        curr_node = Tree(curr_taxid, newrank)
        taxid2node[curr_taxid] = curr_node
        #root
        if curr_taxid == "1":
            curr_node.level_rank = 'R'
            root_node = curr_node
        elif parent_taxid in taxid2node:
            #save parent
            curr_node.parent = taxid2node[parent_taxid]
            curr_node.p_taxid = parent_taxid
            taxid2node[parent_taxid].add_child(curr_node)
        else:
            #parent not linked
            p_notsaved[curr_taxid] = curr_node
            curr_node.p_taxid = parent_taxid 
    nodes_f.close() 
    sys.stdout.write("\r\t%i nodes read\n" % count_nodes)
    sys.stdout.flush()
    #Fix parents
    for taxid in p_notsaved:
        p_taxid = p_notsaved[taxid].p_taxid
        if p_taxid not in taxid2node:
            sys.stderr.write("ERROR: %s not found in nodes.dmp file\n" % p_taxid)
            continue 
        p_node = taxid2node[p_taxid]
        p_notsaved[taxid].parent = p_node
        p_node.add_child(p_notsaved[taxid])
    #STEP 2/5: PARSE SEQID2TAXID FILE TO GET LEAF NODES 
    sys.stdout.write(">> STEP 2/5: Reading %s\n" % args.s2t_file)
    sys.stdout.write("\t%0 taxids read")
    count_taxids = 0
    s2t_f = open(args.s2t_file,'r')
    leaves = {}
    for line in s2t_f:
        taxid = line.strip().split('\t')[-1]
        #taxid not yet saved
        if taxid not in leaves:
            #node exists for taxid
            if taxid in taxid2node:
                leaves[taxid] = taxid2node[taxid]
                count_taxids += 1
                if (count_taxids % 100 == 0):
                    sys.stdout.write("\r\t%i taxids read" % count_taxids)
                    sys.stdout.flush()
            else:
                sys.stderr.write("ERROR: node for %s not found\n" % taxid)
    s2t_f.close()
    sys.stdout.write("\r\t%i taxids read\n" % count_taxids)
    sys.stdout.flush()
    #STEP 3/5: CONDENSE TAXONOMY TO ONLY INCLUDE TAXIDS IN SEQID2TAXID.MAP
    sys.stdout.write(">> STEP 3/5: Condensing taxonomy\n")
    sys.stdout.write("\t%saving 0 taxids")
    count_final = 0
    save_taxids = {} 
    for leaf in leaves:
        if leaf in save_taxids:
            continue
        save_taxids[leaf] = leaves[leaf]
        p_node = taxid2node[leaf].parent
        count_final += 1
        sys.stdout.write("\r\tsaving %i taxids" % count_final)
        sys.stdout.flush()
        while(p_node != None):
            if p_node.taxid not in save_taxids:
                #travel up path to root 
                save_taxids[p_node.taxid] = p_node
                p_node = p_node.parent
                count_final += 1
                sys.stdout.write("\r\tsaving %i taxids" % count_final)
                sys.stdout.flush()
            else: 
                #Parent path already parsed 
                p_node = None
    sys.stdout.write("\r\tsaving %i taxids\n" % count_final)
    sys.stdout.flush()
    #STEP 4/5: PARSE NAMES.DMP TO GET NAMES FOR TAXIDS IN TREE 
    sys.stdout.write(">> STEP 4/5: Reading %s\n" % args.names_file)
    count_names = 0
    sys.stdout.write("\t%i/%i names found" % (count_names, count_final))
    names_f = open(args.names_file,'r')
    for line in names_f:
        [taxid,name] = line.strip().split('\t|\t')[0:2]
        if taxid in save_taxids:
            if save_taxids[taxid].name == '':
                save_taxids[taxid].name = name
                count_names += 1
                sys.stdout.write("\r\t%i/%i names found" % (count_names, count_final))
                sys.stdout.flush()
            elif "scientific name" in line:
                save_taxids[taxid].name = name
    names_f.close()
    sys.stdout.write("\r\t%i/%i names found\n" % (count_names, count_final))
    sys.stdout.flush()
    #STEP 5/5: PRINT NEW TAXONOMY 
    sys.stdout.write(">> STEP 5/5: Printing final taxonomy to %s\n" % args.out_file)
    print_count = 0
    sys.stdout.write("\t%i nodes printed" % print_count)
    sys.stdout.flush()
    o_file = open(args.out_file,'w')
    parse_nodes = [root_node]
    root_node.level_num = 0
    printed_taxids = {"1":root_node}
    while len(parse_nodes) > 0:
        #Get the first node in list
        curr_node = parse_nodes.pop(0)
        print_count += 1
        sys.stdout.write("\r\t%i nodes printed" % print_count)
        sys.stdout.flush()
        #Print current node
        o_file.write("%s\t|\t" % curr_node.taxid)
        if curr_node.taxid == "1":
            o_file.write("1\t|\t")
        else:
            o_file.write("%s\t|\t" % curr_node.p_taxid)
        o_file.write("%s\t|\t" % curr_node.level_rank)
        o_file.write("%s\t|\t" % curr_node.level_num)
        o_file.write("%s\n" % curr_node.name)
        #Parse through children 
        for child in curr_node.children: 
            #Only save taxids needed
            if child.taxid in save_taxids:
                child.level_num = curr_node.level_num + 1
                #Fix children ranks
                if child.level_rank == '-': 
                    if len(curr_node.level_rank) == 1:
                        child.level_rank = curr_node.level_rank + "1"
                    else:
                        new_num = int(curr_node.level_rank[1:]) + 1
                        child.level_rank = curr_node.level_rank[0] + str(new_num)    
                parse_nodes.append(child)
                printed_taxids[child.taxid] = child 
    o_file.close() 
    sys.stdout.write("\r\t%i nodes printed\n" % print_count)
    sys.stdout.flush()
    #Error check
    for taxid in save_taxids:
        if taxid not in printed_taxids:
            sys.stderr.write("ERROR: %s not linked to root\n" % taxid) 
    #End of program
    time = strftime("%m-%d-%Y %H:%M:%S", gmtime())
    sys.stdout.write("PROGRAM END TIME: " + time + '\n')
    sys.exit(0)

#################################################################################
if __name__ == "__main__":
    main()
#################################################################################
##################################END OF PROGRAM#################################
#################################################################################




#!/usr/bin/env python
####################################################################
#kreport2krona.py converts a Kraken-style report into Krona-compatible format
#Copyright (C) 2019-2020 Jennifer Lu, jennifer.lu717@gmail.com

#This file is part of KrakenTools.
#KrakenTools is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 3 of the license, or
#(at your option) any later version.

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of 
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with this program; if not, see <http://www.gnu.org/licenses/>.

####################################################################
#Jennifer Lu, jlu26@jhmi.edu
#Updated: 05/30/2019
#
#This program reads in a Kraken report file and generates
#a Krona-compatible Text report that can be imported into Krona using ktImportText. 
#
#Parameters:
#   -h, --help...............show help message
#   -r X, --report-file X....input kraken report filename 
#   -o X, --output X.........output Krona text filename 
#   --intermediate-ranks.....include non-traditional taxonomic ranks 
#   --no-intermediate-ranks..do not include non-traditional ranks [default] 
#Input file format (tab-delimited)
#   - percentage of total reads
#   - number of reads (including reads within subtree)
#   - number of reads (only at this level)
#   - taxonomic classification of level (U, D, P, C, O, F, G, S,...etc) 
#   - NCBI taxonomic ID 
#   - name of level   
#Output file format (tab-delimited)
#   - Number of reads within subtree of the specified level
#   - Taxonomy tree levels tab-delimited, with level type [k,p,c,o,f,g,s,x]
#
#Methods
#   - main
#   - process_kraken_report
####################################################################
import os, sys, argparse

####################################################################
#process_kraken_report
#usage: parses a single line in the kraken report and extracts relevant information
#input: kraken report file with the following tab delimited lines
#   - percent of total reads
#   - number of reads (including at lower levels)
#   - number of reads (only at this level)
#   - taxonomy classification of level
#       (U, D, P, C, O, F, G, S, -)
#   - taxonomy ID (0 = unclassified, 1 = root, 2 = Bacteria,...etc)
#   - spaces + name
#returns:
#   - classification/genome name
#   - level name (U, -, D, P, C, O, F, G, S)
#   - reads classified at this level and below in the tree
def process_kraken_report(curr_str):
    split_str = curr_str.strip().split('\t')
    if len(split_str) < 2:
        return []
    try:
        int(split_str[1])
    except ValueError:
        return []
    all_reads = int(split_str[1])
    lvl_reads = int(split_str[2])
    level_type = split_str[-3]
    type2main = {'superkingdom':'D','phylum':'P',
        'class':'C','order':'O','family':'F',
        'genus':'G','species':'S'} 
    if len(level_type) > 1:
        if level_type in type2main:
            level_type = type2main[level_type]
        else:
            level_type = '-'
    #Get name and spaces 
    spaces = 0
    name = split_str[-1]
    for char in name:
        if char == ' ':
            name = name[1:]
            spaces += 1
        else:
            break
    name = name.replace(' ','_')
    #Determine level based on number of spaces
    level_num = spaces/2
    return [name, level_num, level_type, lvl_reads]

###################################################################
#kreport2krona_all
#usage: prints all levels for a kraken report 
#input: kraken report file and output krona file names 
#returns: none 
def kreport2krona_all(report_file, out_file):
    #Process report file and output 
    curr_path = [] 
    prev_lvl_num = -1
    r_file = open(report_file, 'r')
    o_file = open(out_file, 'w')
    #Read through report file 
    main_lvls = ['D','P','C','O','F','G','S']
    for line in r_file:
        report_vals = process_kraken_report(line)
        #If header line, skip
        if len(report_vals) < 4: 
            continue
        #Get relevant information from the line 
        [name, level_num, level_type, lvl_reads] = report_vals
        if level_type == 'U':
            o_file.write(str(lvl_reads) + "\tUnclassified\n")
            continue
        #Create level name 
        if level_type not in main_lvls:
            level_type = "x"
        elif level_type == "D":
            level_type = "K"
        level_str = level_type.lower() + "__" + name
        #Determine full string to add
        if prev_lvl_num == -1:
            #First level
            prev_lvl_num = level_num
            curr_path.append(level_str)
            o_file.write(str(lvl_reads) + "\t" + level_str + "\n")
        else:
            o_file.write(str(lvl_reads))
            #Move back if needed
            while level_num != (prev_lvl_num + 1):
                prev_lvl_num -= 1
                curr_path.pop()
            #Print all ancestors of current level followed by |
            for string in curr_path:
                if string[0] != "r": 
                    o_file.write("\t" + string)
            #Print final level and then number of reads
            o_file.write("\t" + level_str + "\n")
            #Update
            curr_path.append(level_str)
            prev_lvl_num = level_num
    o_file.close()
    r_file.close()
    
###################################################################
#kreport2krona_main
#usage: prints only main taxonomy levels for a kraken report 
#input: kraken report file and output krona file names 
#returns: none 
def kreport2krona_main(report_file, out_file):
    #Process report file and output 
    main_lvls = ['D','P','C','O','F','G','S']
    curr_path = [] 
    prev_lvl_num = -1
    num2path = {} 
    path2reads = {} 
    line_num = -1
    #Read through report file 
    r_file = open(report_file, 'r')
    for line in r_file:
        line_num += 1
        #########################################
        report_vals = process_kraken_report(line)
        #If header line, skip
        if len(report_vals) < 4: 
            continue
        #Get relevant information from the line 
        [name, level_num, level_type, lvl_reads] = report_vals
        if level_type == 'U':
            num2path[line_num] = ["Unclassified"]
            path2reads["Unclassified"] = lvl_reads 
            continue
        #########################################
        #Create level name 
        if level_type not in main_lvls:
            level_type = "x"
        elif level_type == "D":
            level_type = "K"
        level_str = level_type.lower() + "__" + name
        #########################################
        #Determine full string to add
        if prev_lvl_num == -1:
            #First level
            prev_lvl_num = level_num
            curr_path.append(level_str)
            #Save
            if curr_path[-1][0] == "x":
                num2path[line_num] = ""
            else:
                path2reads[curr_path[-1]] = lvl_reads
                num2path[line_num] = []
                for i in curr_path:
                    num2path[line_num].append(i)
            continue
        else:
            #########################################
            #Move back if needed
            while level_num != (prev_lvl_num + 1):
                prev_lvl_num -= 1
                curr_path.pop()
            #Update the list 
            curr_path.append(level_str)
            prev_lvl_num = level_num
            #########################################
            #IF AT NON-TRADITIONAL LEVEL, ADD TO PARENT
            if level_type == "x":
                test_num = len(curr_path) - 1
                while(test_num >= 0):
                    if curr_path[test_num][0] != "x":
                        path2reads[curr_path[test_num]] += lvl_reads 
                        test_num = -1
                    test_num = test_num - 1 
                num2path[line_num] = ""
            #IF AT TRADITIONAL LEVEL, SAVE 
            if level_type != "x":
                path2reads[curr_path[-1]] = lvl_reads
                num2path[line_num] = []
                for i in curr_path:
                    num2path[line_num].append(i)
    r_file.close() 
    
    #WRITE OUTPUT FILE
    o_file = open(out_file, 'w')
    for i in range(0,line_num+1):
        #Get values
        if i not in num2path:
            continue
        curr_path = num2path[i] 
        if len(curr_path) > 0:
            curr_reads = path2reads[curr_path[-1]] 
            if curr_path[-1][0] != "x":
                o_file.write("%i" % curr_reads)
            for name in curr_path:
                if name[0] != "r" and name[0] != "x":
                    o_file.write("\t%s" % name)
            o_file.write("\n")
    o_file.close()

######################################################################
#Main method
def main():
    #Parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('-r', '--report-file', '--report', required=True,
        dest='r_file', help='Input kraken report file for converting')
    parser.add_argument('-o', '--output', required=True,
        dest='o_file', help='Output krona-report file name')
    parser.add_argument('--intermediate-ranks', action='store_true',
        dest='x_include', default=False, required=False,
        help='Include non-traditional taxonomic ranks in output')
    parser.add_argument('--no-intermediate-ranks', action='store_false',
        dest='x_include', default=False, required=False,
        help='Do not include non-traditional taxonomic ranks in output [default: no intermediate ranks]')
    args=parser.parse_args()

    #Determine which krona report to make 
    if args.x_include:
        kreport2krona_all(args.r_file,args.o_file)
    else:
        kreport2krona_main(args.r_file,args.o_file) 

#################################################################
if __name__ == "__main__":
    main()
#########################END OF PROGRAM##########################



#!/usr/bin/env python
import os
import sys
import argparse
import gzip
from time import gmtime, strftime
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

class Tree(object):
    """Tree node used in constructing taxonomy tree."""
    def __init__(self, taxid, level_num, level_id, children=None, parent=None):
        self.taxid = taxid
        self.level_num = level_num
        self.level_id = level_id
        self.children = []
        self.parent = parent
        if children is not None:
            for child in children:
                self.add_child(child)
                
    def add_child(self, node):
        assert isinstance(node, Tree)
        self.children.append(node)

def process_kraken_output(kraken_line):
    l_vals = kraken_line.split('\t')
    if len(l_vals) < 5:
        return [-1, '']
    tax_id = l_vals[2].split("taxid ")[-1] if "taxid" in l_vals[2] else l_vals[2]
    tax_id = 81077 if tax_id == 'A' else int(tax_id)
    return [tax_id, l_vals[1]]

def process_kraken_report(report_line):
    l_vals = report_line.strip().split('\t')
    if len(l_vals) < 5:
        return []
    try:
        int(l_vals[1])
    except ValueError:
        return []
    taxid = int(l_vals[-3])
    level_type = l_vals[-2]
    map_kuniq = {'species': 'S', 'genus': 'G', 'family': 'F', 'order': 'O', 'class': 'C', 'phylum': 'P', 'superkingdom': 'D', 'kingdom': 'K'}
    level_type = map_kuniq.get(level_type, '-')
    level_num = int(sum(1 for char in l_vals[-1] if char == ' ') / 2)
    return [taxid, level_num, level_type]

def extract_reads_i(args):
    # Your program logic using args (parsed arguments passed as function arguments)
    # Initialize taxonomies, open files, and execute the core read extraction logic
    # -- Here you would keep the main logic of the program without direct argument parsing.
    pass

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-k', dest='kraken_file', required=True, help='Kraken output file to parse')
    parser.add_argument('-s','-s1', '-1', '-U', dest='seq_file1', required=True, help='FASTA/FASTQ File containing the raw sequence letters.')
    parser.add_argument('-s2', '-2', dest='seq_file2', default="", help='2nd FASTA/FASTQ File containing the raw sequence letters (paired).')
    parser.add_argument('-t', "--taxid", dest='taxid', required=True, nargs='+', help='Taxonomy ID[s] of reads to extract (space-delimited)')
    parser.add_argument('-o', "--output", dest='output_file', required=True, help='Output FASTA/Q file containing the reads and sample IDs')
    parser.add_argument('-o2', "--output2", dest='output_file2', required=False, default='', help='Output FASTA/Q file containing the second pair of reads [required for paired input]')
    parser.add_argument('--append', dest='append', action='store_true', help='Append the sequences to the end of the output FASTA file specified.')
    parser.add_argument('--noappend', dest='append', action='store_false', help='Create a new FASTA file (rewrite if existing) [default].')
    parser.add_argument('--max', dest='max_reads', required=False, default=100000000, type=int, help='Maximum number of reads to save [default: 100,000,000]')
    parser.add_argument('-r', '--report', dest='report_file', required=False, default="", help='Kraken report file. [required only if --include-parents/children is specified]')
    parser.add_argument('--include-parents', dest="parents", required=False, action='store_true', default=False, help='Include reads classified at parent levels of the specified taxids')
    parser.add_argument('--include-children', dest='children', required=False, action='store_true', default=False, help='Include reads classified more specifically than the specified taxids')
    parser.add_argument('--exclude', dest='exclude', required=False, action='store_true', default=False, help='Finds all reads NOT matching specified taxids')
    parser.add_argument('--fastq-output', dest='fastq_out', required=False, action='store_true', default=False, help='Print output FASTQ reads [default: output is FASTA]')
    parser.set_defaults(append=False)

    args = parser.parse_args()
    extract_reads_i(args)

if __name__ == "__main__":
    main()



import os
import gzip
from multiprocessing import Pool
from itmfinder.itm_helper.extract_kraken_reads import extract_reads_i
from itmfinder.utils.data_loader import load_contamination_list

def print_colorful_message(message, color):
    """
    Print a colorful message to the console.

    Args:
        message (str): The message to be printed.
        color (str): The color code to be applied.
                     'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white'
    """
    colors = {
        'red': '\033[91m',
        'green': '\033[92m',
        'yellow': '\033[93m',
        'blue': '\033[94m',
        'magenta': '\033[95m',
        'cyan': '\033[96m',
        'white': '\033[97m',
    }
    end_color = '\033[0m'
    if color not in colors:
        print("Invalid color. Please choose from 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white'.")
        return
    colored_message = f"{colors[color]}{message}{end_color}"
    print(colored_message)


def compress_to_gz(file_path):
    if file_path and os.path.exists(file_path):
        with open(file_path, 'rb') as f_in:
            with gzip.open(file_path + '.gz', 'wb') as f_out:
                f_out.writelines(f_in)
        # After compression, delete the original file
        os.remove(file_path)
    else:
        print(f"File {file_path} not found or invalid. Compression skipped.")

def rmc_extract_microbiome_reads(sample_info, is_single_end=False, itm_path=None):
    """
    Extract microbiome reads for a given sample.

    Args:
        sample_info (tuple): A tuple containing sample information.
        is_single_end (bool): Whether to process single-end data. Default is False.
        itm_path (str): Path to ITMfinder. Default is None.
    """
    sample_id, fastq_r1, kraken_output, output_r1, output_r3, output_r4, path4_ku1, path6_rcr = sample_info
    report_file = os.path.join(path4_ku1, f"{sample_id}.kraken.report.txt")
    task_complete_file = os.path.join(path6_rcr, f"{sample_id}.task.complete")

    # Check if output files and task_complete file already exist
    if os.path.exists(report_file) and os.path.exists(task_complete_file):
        print(f">>> Sample {sample_id} already processed. Skipping...")
        return

    # conta_file = os.path.join(itm_path, "itm_helper", "conta_list.txt")
    # with open(conta_file, "r") as file:
    #     lines = file.readlines()
    # second_column = [line.strip().split()[1] for line in lines if len(line.strip().split()) > 1]
    
    # Filter out any non-integer values
    # conta_ls = [taxid for taxid in second_column if taxid.isdigit()]
    
    contamination_ids = load_contamination_list()

    print_colorful_message(f">>> Running decontamination process for sample {sample_id}...", "cyan")

    if is_single_end:
        # process_args = ["python", os.path.join(itm_path, "itm_helper", "extract_kraken_reads.py"),
        #                 "-k", kraken_output, "-U", fastq_r1, "-o", output_r1,
        #                 "--taxid", *conta_ls, "--exclude", "--include-children", "-r", report_file]
        extract_reads_i(kraken_file=kraken_output, seq_file1=fastq_r1, output_file1=output_r1, 
                        taxid=contamination_ids, report_file=report_file, children=True, exclude=True)
    else:
        fastq_r2 = fastq_r1.replace("_mr_1.fastq.gz", "_mr_2.fastq.gz")

        extract_reads_i(kraken_file=kraken_output, seq_file1=fastq_r1, seq_file2=fastq_r2, output_file1=output_r3, 
                        output_file2=output_r4, taxid=contamination_ids, report_file=report_file, children=True, exclude=True)

        # process_args = ["python", os.path.join(itm_path, "itm_helper", "extract_kraken_reads.py"),
        #                 "-k", kraken_output, "-s1", fastq_r1, "-s2", fastq_r2, "-o", output_r3, "-o2", output_r4,
        #                 "--taxid", *conta_ls, "--exclude", "--include-children", "-r", report_file]

    # Check if all required arguments are not None
    process_args = [arg for arg in process_args if arg is not None]

    process = subprocess.Popen(process_args)
  
    process.wait()
    if process.returncode == 0:
        with open(task_complete_file, "w") as f:
            f.write("Task completed.")
        print(f">>>--- Sample {sample_id} processed successfully.")
        print("   ")

        # compress output files if they are not None
        if is_single_end:
            compress_to_gz(output_r1)
        else:
            compress_to_gz(output_r3)
            compress_to_gz(output_r4)
    else:
        print(f"Error occurred while processing sample {sample_id}.")


def step5_decontamination(path5_mr, path4_ku1, path6_rcr, itm_path, batch_size=1, is_single_end=False):
    """
    Extract microbiome reads from Kraken2 output.

    Args:
        path5_mr (str): Path to host removed FASTQ files.
        path4_ku1 (str): Path to Kraken2 outputs.
        path6_rcr (str): Path to extracted microbiome reads.
        itm_path (str): Path to ITMfinder.
        batch_size (int): Number of samples to process simultaneously. Default is 1.
        is_single_end (bool): Whether to process single-end data. Default is False.
    """

    print("   ")
    print_colorful_message("#########################################################", "blue")
    print_colorful_message(" ITMfinder: Identifing Intratumoral Microbiome pipeline ", "cyan")
    print_colorful_message(" If you encounter any issues, please report them at ", "cyan")
    print_colorful_message(" https://github.com/LiaoWJLab/ITMfinder/issues ", "cyan")
    print_colorful_message("#########################################################", "blue")
    print(" Author: Dongqiang Zeng, Qianqian Mao ")
    print(" Email: interlaken@smu.edu.cn ")
    print_colorful_message("#########################################################", "blue")
    print("   ")

    # Notify the user about using extract_kraken_reads.py from KrakenTools
    print("  Extracting microbiome reads using extract_kraken_reads.py from KrakenTools:")
    print("  https://github.com/jenniferlu717/KrakenTools")

    print("   ")
    print(" >>> Step-5 Decontaminate Microbiome Reads...  ")

    # Create the output directory if it does not exist
    os.makedirs(path6_rcr, exist_ok=True)

    # Get list of forward reads files
    if is_single_end:
        fastq_files = [file for file in os.listdir(path5_mr) if file.endswith("_mr.fastq.gz")]
    else:
        fastq_files = [file for file in os.listdir(path5_mr) if file.endswith("_mr_1.fastq.gz")]

    # Construct sample information tuples
    samples = []
    for fastq_r1 in fastq_files:
        if is_single_end:
            sample_id = fastq_r1.replace("_mr.fastq.gz", "")
            fastq_r1_path = os.path.join(path5_mr, f"{sample_id}_mr.fastq.gz")
            samples.append((sample_id, fastq_r1_path, os.path.join(path4_ku1, f"{sample_id}.kraken.output.txt"),
                            os.path.join(path6_rcr, f"{sample_id}_rcr.fastq"), None, None, path4_ku1, path6_rcr))
        else:
            sample_id = fastq_r1.replace("_mr_1.fastq.gz", "")
            fastq_r1_path = os.path.join(path5_mr, f"{sample_id}_mr_1.fastq.gz")
            fastq_r2_path = os.path.join(path5_mr, f"{sample_id}_mr_2.fastq.gz")
            samples.append((sample_id, fastq_r1_path, os.path.join(path4_ku1, f"{sample_id}.kraken.output.txt"), None, 
                            os.path.join(path6_rcr, f"{sample_id}_rcr_1.fastq"), os.path.join(path6_rcr, f"{sample_id}_rcr_2.fastq"), 
                            path4_ku1, path6_rcr))

    # Process samples in parallel using multiprocessing Pool
    with Pool() as pool:
        for i in range(0, len(samples), batch_size):
            if is_single_end:
                pool.starmap(rmc_extract_microbiome_reads, [(sample, is_single_end, itm_path) for sample in samples[i:i + batch_size]])
            else:
                pool.starmap(rmc_extract_microbiome_reads, [(sample, is_single_end, itm_path) for sample in samples[i:i + batch_size]])
    print("   ")
    print(">>>=== Microbiome reads extraction completed.")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Step 5: Decontaminate microbiome reads from Kraken2 output")
    parser.add_argument("--path5_mr", type=str, help="Path to extracted microbiome reads")
    parser.add_argument("--path4_ku1", type=str, help="Path to Kraken2 outputs")
    parser.add_argument("--path6_rcr", type=str, help="Path to decontaminat microbiome reads")
    parser.add_argument("--itm_path", type=str, help="Path to ITMfinder")
    parser.add_argument("--batch_size", type=int, default=1, help="Number of samples to process simultaneously")
    parser.add_argument("--se", action="store_true", help="Use single-end processing. Default is paired-end.")
    args = parser.parse_args()

    step5_decontamination(args.path5_mr, args.path4_ku1, args.path6_rcr, args.itm_path, args.batch_size, is_single_end=args.se)



from setuptools import setup, find_packages

setup(
    name="itmfinder",
    version="0.1.1",
    packages=find_packages(),
    # packages=find_packages(),  # 自动发现并包含所有包
    include_package_data=True,
    package_data={
        'itmfinder': ['resources/*'],  # 这样包含resources目录下的文件
    },
    entry_points={
        'console_scripts': [
            'itmfinder=itmfinder.main:main',   # 创建一个命令行工具链接到 main.py 中的 main 函数
        ],
    },
    install_requires=[
        'numpy',
        'pandas',
        'biopython',
        'matplotlib',
        'scikit-learn',
    ],
    author="Dongqiang Zeng",
    author_email='interlaken@smu.edu.cn',
    license='MIT License',
    description="Intratumoral Microbiome Finder Tool",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/DongqiangZeng0808/itmfinder",
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



#!/usr/bin/env python
################################################################
#bray_curtis.py takes multiple reports of classifications or
#abundances and calculates a matrix of bray-curtis dissimilarity
#metrics
#Copyright (C) 2019 Jennifer Lu, jlu26@jhmi.edu
#
#This file is part of KrakenTools
#KrakenTools is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 3 of the license, or
#(at your option) any later version.

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program; if not, see <http://www.gnu.org/licenses/>.

#################################################################
#Jennifer Lu, jlu26@jhmi.edu
#Christopher Pockrandt, pockrandt@jhu.edu
#Updated: 08/12/2019
#
#This program takes multiple reports of classifications or
#abundances and calculates a matrix of bray-curtis dissimilarity
#metrics
#
#Parameters:
#   -h, --help................show help message.
#   -i X, --input-files X.....all (at least 2) input files (separated by spaces)
#Input File options (all input files must be of the same format)
#   --single [default]........all samples are within a single tab-delimited file
#                             where the first column is the category and the rest are counts
#   --simple .................input files must be tab-delimited with the first
#                             two columns representing categories and then counts
#   --bracken.................input files are Bracken outputs: col  2=taxonomy, 6=counts
#   --kraken..................input files are Kraken reports: col 5=taxonomy, 3=counts
#   --krona...................input files are Krona format: col 1=counts, 2=taxonomy
#Input options:
#   --level [S, G, etc].......user specifies which level to measure at
#                             (for kraken, krona, or bracken input files)
####################################################################
import os, sys, argparse
import operator
from time import gmtime
from time import strftime
import numpy as np
####################################################################
#Main method
def main():
    #Parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('-i','--input','--input-files',
        '--inputs', required=True,dest='in_files',nargs='+',
        help='Input files (one per community) for which to compare for \
        bray-curtis dissimiliarity metrics')
    parser.add_argument('--type', required=False, default='single', dest='filetype',
        choices=['single','simple','bracken','kreport','kreport2','krona'],
        help='Type of input file[s]: single, simple [tab-delimited, specify --cols], \
            bracken, kreport, kreport2, krona. See docs for details')
    parser.add_argument('--cols','--columns', dest='cols',required=False, default='1,2',
        help='Specify category/counts separated by single comma: cat,counts (1 = first col)')
    parser.add_argument('--level', '-l', dest='lvl', required=False, default='all',choices=['all', 'S', 'G', 'F', 'O'],
        help='For Kraken or Krona files, taxonomy level for which to compare samples. Default: all')
    args=parser.parse_args()

    #################################################
    #Test input files
    in2counts = {}
    if args.filetype == 'single' and len(args.in_files) > 1:
        sys.stderr.write("Please specify only one file for '--type simple'\n")
        exit(1)
    for f in args.in_files:
        if not os.path.isfile(f):
            sys.stderr.write("File %s not found\n" % f)
            exit(1)

    #################################################
    #Determine columns for extracting
    categ_col = -1
    count_col = -1
    if args.filetype in ['single','simple']:
        if ',' not in args.cols:
            sys.stderr.write("Please specify column as 'a,b' where a = column of category, \
            b = column of first count\n")
            exit(1)
        else:
            [categ_col, count_col] = args.cols.split(',')
            if not categ_col.isdigit():
                sys.stderr.write("%s is not an integer\n" % categ_col)
                exit(1)
            elif not count_col.isdigit():
                sys.stderr.write("%s is not an integer\n" % count_col)
                exit(1)
            categ_col = int(categ_col) - 1
            count_col = int(count_col) - 1
    elif args.filetype == "bracken":
        categ_col = 0 # TODO taxid (col 1) does not seem to be properly set
        count_col = 5
        taxlvl_col = 2
    elif args.filetype == "kreport" or args.filetype == "kreport2": # TODO: what about kuniq reports?
        categ_col = 4
        count_col = 2
        taxlvl_col = 3
    elif args.filetype == "krona":
        categ_col = 1
        count_col = 0
    #################################################
    #STEP 1: READ IN SAMPLES
    i2totals = {}
    i2counts = {}
    i2names = {}
    num_samples = 0
    num_categories = 0
    if args.filetype == "single":
        #ALL SAMPLE COUNTS WITHIN A SINGLE FILE
        #sys.stdout.write(">>STEP 1: READING INPUT FILE\n")
        #sys.stdout.write("\t....reading line 1 as header\n")
        header = True
        i_file = open(args.in_files[0],'r')
        for line in i_file:
            l_vals = line.strip().split("\t")
            #Read header
            if header:
                s_count = 0
                for i in range(count_col, len(l_vals)):
                    i2names[s_count] = l_vals[i]
                    i2counts[s_count] = {}
                    i2totals[s_count] = 0
                    s_count += 1
                num_samples = s_count
                header = False
                #sys.stdout.write("\t....reading %i samples\n" % s_count)
            else:
                #Otherwise, save_counts
                s_count = 0
                curr_categ = l_vals[categ_col]
                for i in range(count_col, len(l_vals)):
                    if int(l_vals[i]) > 0:
                        i2totals[s_count] += int(l_vals[i])
                        i2counts[s_count][curr_categ] = int(l_vals[i])
                    s_count += 1
                num_categories += 1
        i_file.close()
        #sys.stdout.write("\t....finished reading counts for %i samples in %i categories\n" % (num_samples,num_categories))
    else: # for braken, kraken, kraken2 and krona
        num_samples = 0
        i2names = {}
        i2totals = {}
        i2counts = {}
        genus = {}

        for f in args.in_files:
            i_file = open(f,'r')
            i2names[num_samples] = f
            i2totals[num_samples] = 0
            i2counts[num_samples] = {}
            genus[num_samples] = {}

            for line in i_file:
                l_vals = line.strip().split("\t")

                # empty line, header line or line is a comment
                if len(l_vals) == 0 or (not l_vals[count_col].isdigit()) or l_vals[0] == '#':
                    continue

                if int(l_vals[count_col]) > 0:
                    if args.filetype == "krona":
                        # we don't know which column will be the genus type (might change for every line)
                        # hence, we iterate over it
                        for i in range(count_col, len(l_vals)):
                            if l_vals[i].startswith(args.lvl.lower() + "__") or args.lvl == "all":
                                categ_col = i
                                tax_name = l_vals[categ_col]
                                i2totals[num_samples] += int(l_vals[count_col])
                                if not tax_name in i2counts[num_samples]:
                                    i2counts[num_samples][tax_name] = 0
                                i2counts[num_samples][tax_name] += int(l_vals[count_col])
                                # sys.stdout.write("%s\t%s\t%i\n" % (f, tax_name, int(l_vals[count_col])))
                    else:
                        if l_vals[taxlvl_col][0] == args.lvl or args.lvl == "all": # TODO: what if it is G1?
                            # TODO: cant do this because of broken bracken files: tax_id = int(l_vals[categ_col])
                            tax_id = l_vals[categ_col]
                            genus[num_samples][tax_id] = l_vals[0]
                            i2totals[num_samples] += int(l_vals[count_col])
                            if tax_id not in i2counts[num_samples]:
                                i2counts[num_samples][tax_id] = 0
                            i2counts[num_samples][tax_id] += int(l_vals[count_col])
                            # sys.stdout.write("%s\t%s\t%i\n" % (f, line, i2counts[num_samples][tax_id]))
            i_file.close()
            num_samples += 1
    #################################################
    #STEP 2: CALCULATE BRAY-CURTIS DISSIMILARITIES
    #sys.stdout.write(">>STEP 2: COMPARING SAMPLES TO CALCULATE DISSIMILARITIES\n")

    bc = np.zeros((num_samples,num_samples))
    for i in range(0,num_samples):
        i_tot = i2totals[i]
        for j in range(i+1, num_samples):
            j_tot = i2totals[j]
            C_ij = 0.0
            for cat in i2counts[i]:
                if cat in i2counts[j]:
                    C_ij += min(i2counts[i][cat], i2counts[j][cat])
            #Calculate bray-curtis dissimilarity
            bc_ij = 1.0 - ((2.0*C_ij)/float(i_tot+j_tot))
            #sys.stdout.write("%s\t%s\t%i\t%i\t%i\n" % (i2names[i], i2names[j], C_ij, i_tot, j_tot))
            bc[i][j] = bc_ij
            bc[j][i] = bc_ij

    # for i in range(0,num_samples):
    #     sys.stdout.write("Totals: %s\t%i\n" % (i2names[i], i2totals[i]))
    #     for cat in i2counts[i]:
    #         sys.stdout.write("%i\t%s\t%i\n" % (cat, genus[i][cat], i2counts[i][cat]))

    #################################################
    #sys.stdout.write(">>STEP 3: PRINTING MATRIX OF BRAY_CURTIS DISSIMILARITIES\n")
    #sys.stdout.flush()
    #Print samples
    for i in i2names:
        sys.stdout.write("#%i\t%s (%i reads)\n" % (i,i2names[i],i2totals[i]))
    #Print headers
    sys.stdout.write("x")
    for i in range(num_samples):
        sys.stdout.write("\t%i" % i)
    sys.stdout.write("\n")
    #Print matrix
    for i in range(num_samples):
        sys.stdout.write("%i" % i)
        for j in range(num_samples):
            if i <= j:
                sys.stdout.write("\t%0.3f" % bc[i][j])
            else:
                sys.stdout.write("\tx.xxx")
        sys.stdout.write("\n")

####################################################################
if __name__ == "__main__":
    main()



#!/usr/bin/env python
import os, sys, argparse
import math

def shannons_alpha(p):
	# shannons = -sum(pi ln(pi))
	h = []
	for i in p:
		h.append(i * math.log(i))
	print("Shannon's diversity: %s" %(-1 *sum(h)))
	return (-1 *sum(h))

def berger_parkers_alpha(p):
	# bp is nmax/N which is equal to the max pi == max(p)
	print("Berger-parker's diversity: %s" %max(p))
	return max(p)

def simpsons_alpha(D):
	# simpsons index of diversity = 1 - D
	# D = (sum ni(ni-1))/ N*(N-1)
	print("Simpson's index of diversity: %s" %(1-D))
	return 1-D

def inverse_simpsons_alpha(D):
	# simsons inverse = 1/D
	print("Simpson's Reciprocal Index: %s" %(1/D))
	return 1/D

def fishers_alpha():	
	global np
	import numpy as np
	from scipy.optimize import fsolve
	
	fish = fsolve(eqn_output,1)
	
	# NOTE:
	# if ratio of N/S > 20 then x > 0.99 (Poole,1974)
	# x is almost always > 0.9 and never > 1.0 i.e. ~ 0.9 < x < 1.0
	print("Fisher's index: %s" %fish[0])
	return fish

def eqn_output(a):
	return a * np.log(1+N_f/a) - S_f

# Main method
def main():
	# get arguments
	parser = argparse.ArgumentParser(description='Pick an alpha diversity.')
	parser.add_argument('-f','--filename',dest='filename',help='bracken file with species abundance estimates')
	parser.add_argument('-a','--alpha',dest='value',default='Sh',type=str,  help='type of alpha diversity to calculate Sh, BP, Si, ISi, F, default = Sh')
	args = parser.parse_args()

	f = open(args.filename)
	
	f.readline()
	n = []
	# read in the file
	for line in f: 
		ind_abund = line.split('\t')[5] # finds the abundance estimate
		n.append(float(ind_abund)) 
	
	f.close()
	
	# calculations
	N = sum(n) # total number of individuals
	S = len(n) # total number of species
	# calculate all the pi's
	p = [] # store pi's 
	D = 0
	for i in n: # go through each species
		if i != 0: # there should not be any zeros
			p.append(i/N) # pi is the ni/N
			D += i*(i-1)
	
	D = D/(N*(N-1))
	# find the indicated alpha
	if args.value == 'Sh': # calculate shannon's diversity
		shannons_alpha(p)
	elif args.value == 'BP': # calculate berger-parker's dominance index
		berger_parkers_alpha(p)
	elif args.value == 'Si': # calculate Simpson's alpha 
		simpsons_alpha(D)
	elif args.value == 'ISi': # calculate Inverse Simpson's alpha 
		inverse_simpsons_alpha(D)
	elif args.value == 'F': # calculate fisher's alpha
		print("Fisher's alpha...loading")
		global N_f
		N_f = sum(n)
		global S_f
		S_f = len(n)
		fishers_alpha()
	else:
		print("Not a supported alpha")

if __name__ == "__main__":
    main()




# -*- coding: utf-8 -*-
"""
Created on Mon Apr 17 17:15:36 2023

@author: sipost1
"""
import os
import glob
import csv
import argparse
from collections import defaultdict
from typing import Dict, TextIO

def extract(file: str) -> Dict:
    """
    Requires a valid kraken2 report file.
    Returns a dictionary of taxa and read counts.
    """
    taxons = defaultdict(dict)

    # Standard kraken2 output has 6 fields.
    indexes = {6: [3, 5], 8: [5, 7]}
    file_path = os.path.abspath(file)

    with open(file_path, "r") as ori:
        lines = ori.readlines()

    assert len(lines) > 0, "Empty file!"

    taxon_index, name_index = indexes[len(lines[0].split("\t"))]
    print(f"Reading {file}")
    print(f"Used indexes: field {taxon_index + 1} for taxon rank, field {name_index + 1} for taxon name.")

    for line in lines:
        line_params = line.rstrip("\n").split("\t")
        read_count = line_params[1]
        taxon = line_params[taxon_index].strip()
        name = line_params[name_index].strip()
        taxons[taxon][name] = read_count

    return taxons

def read_in_files(directory: str, extension: str = ".k2report") -> Dict:
    """
    Reads k2report files from a directory and returns a 3-depth dictionary.
    """
    file_dictionary = defaultdict()
    report_files = glob.glob(f"{directory}/*{extension}")
    assert len(report_files) > 0, "No report file found! Please check filename extension and input directory!"

    for file in report_files:
        file = os.path.abspath(file)
        file_dictionary[os.path.basename(file).rstrip(extension)] = extract(file)

    return file_dictionary

def create_otu_table(level: str, file_sample_dict: Dict, outdir: str = "./") -> TextIO:
    """
    Creates an OTU table for a given taxonomic level.
    """
    rearranged_dict = defaultdict(dict)

    level = level.upper()
    sample_taxa = {sample: level_list[level] for sample, level_list in file_sample_dict.items()}

    for sample, taxon_list in sample_taxa.items():
        for taxon in taxon_list:
            rearranged_dict[taxon][sample] = taxon_list[taxon]

    headers = ["otu"] + list(sample_taxa.keys())
    outfile_name = f"otu_table_{level}.csv"
    with open(os.path.join(outdir, outfile_name), "w", newline="") as csv_file:
        writer = csv.writer(csv_file)
        writer.writerow(headers)  # Write headers

        for otu, inner_dict in rearranged_dict.items():
            row = [otu] + [inner_dict.get(header, 0) for header in headers[1:]]
            writer.writerow(row)

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--inputfolder", "-i", type=str, required=True, help="Input folder where report files can be found.")
    parser.add_argument("--level", "-l", type=str, required=True, help="Taxonomic level to extract from kraken2 report.")
    parser.add_argument("--extension", "-e", default=".k2report", help="Extension of kraken2 report files. Default: .k2report")
    parser.add_argument("--outdir", "-o", type=str, default="./", help="Output directory to save OTU table files. Default: current directory")
    args = parser.parse_args()

    # Process files and generate OTU table
    file_dict = read_in_files(args.inputfolder, extension=args.extension)
    create_otu_table(args.level, file_dict, outdir=args.outdir)
    print("Done!")

# Only execute main if this script is run directly
if __name__ == "__main__":
    main()



#!/usr/bin/env python
####################################################################
#kreport2mpa.py converts a Kraken-style report into mpa [MetaPhlAn) format
#Copyright (C) 2017-2020 Jennifer Lu, jennifer.lu717@gmail.com

#This file is part of KrakenTools.
#KrakenTools is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 3 of the license, or
#(at your option) any later version.

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of 
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with this program; if not, see <http://www.gnu.org/licenses/>.

####################################################################
#Jennifer Lu, jlu26@jhmi.edu
#11/06/2017
#Updated: 07/12/2020
#
#This program reads in a Kraken report file and generates
#an mpa-format (MetaPhlAn) style report. Each line represents
#a possible taxon classification. The first column is lists the 
#domain, kingdom, phyla, etc, leading up to each taxon.
#The levels are separated by the | delimiter, with the type of 
#level specified before each name with a single letter and underscore
#(d_ for domain, k_ for kingdom, etc). 
#The second column is the number of reads classified within 
#that taxon's subtree.
#
#Input file:
#   - Kraken report file generates from the kraken raw output file
#Input Parameters to Specify [OPTIONAL]:
#   - header_line = prints a header line in mpa-report 
#       [Default: no header]
#   - intermediate-ranks = includes non-traditional taxon levels
#       (traditional levels: domain, kingdom, phylum, class, order, 
#       family, genus, species)
#       [Default: no intermediate ranks]
#Output file format (tab-delimited)
#   - Taxonomy tree levels |-delimited, with level type [d,k,p,c,o,f,g,s,x]
#   - Number of reads within subtree of the specified level
#
#Methods
#   - main
#   - process_kraken_report
#
#!/usr/bin/env python
import os
import argparse

def process_kraken_report(curr_str, remove_spaces):
    split_str = curr_str.strip().split('\t')
    if len(split_str) < 4:
        return []
    try:
        int(split_str[1])
    except ValueError:
        return []
    percents = float(split_str[0])
    all_reads = int(split_str[1])
    try:
        taxid = int(split_str[-3]) 
        level_type = split_str[-2]
        map_kuniq = {'species':'S', 'genus':'G','family':'F', 'order':'O', 'class':'C', 'phylum':'P', 'superkingdom':'D', 'kingdom':'K'}
        level_type = map_kuniq.get(level_type, '-')
    except ValueError:
        taxid = int(split_str[-2])
        level_type = split_str[-3]

    spaces = 0
    name = split_str[-1]
    while name.startswith(' '):
        name = name[1:]
        spaces += 1
    if remove_spaces:
        name = name.replace(' ', '_')
    level_num = spaces // 2
    return [name, level_num, level_type, all_reads, percents]

def kreport2mpai(r_file_path, o_file_path, add_header=False, use_reads=True, x_include=False, remove_spaces=True):
    curr_path = [] 
    prev_lvl_num = -1
    with open(r_file_path, 'r') as r_file, open(o_file_path, 'w') as o_file:
        if add_header:
            o_file.write("#Classification\t" + os.path.basename(r_file_path) + "\n")
        
        main_lvls = ['R', 'K', 'D', 'P', 'C', 'O', 'F', 'G', 'S']
        for line in r_file:
            report_vals = process_kraken_report(line, remove_spaces)
            if len(report_vals) < 5: 
                continue
            [name, level_num, level_type, all_reads, percents] = report_vals
            if level_type == 'U':
                continue
            level_type = "x" if level_type not in main_lvls else level_type.lower() if level_type == "D" else level_type
            level_str = level_type + "__" + name

            if prev_lvl_num == -1:
                prev_lvl_num = level_num
                curr_path.append(level_str)
            else:
                while level_num != (prev_lvl_num + 1):
                    prev_lvl_num -= 1
                    curr_path.pop()

                if (level_type == "x" and x_include) or level_type != "x":
                    for string in curr_path:
                        if (string[0] == "x" and x_include) or string[0] != "x":
                            if string[0] != "r": 
                                o_file.write(string + "|")
                    o_file.write(level_str + "\t" + str(all_reads if use_reads else percents) + "\n")
                curr_path.append(level_str)
                prev_lvl_num = level_num

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-r', '--report-file', '--report', required=True, dest='r_file', help='Input kraken report file for converting')
    parser.add_argument('-o', '--output', required=True, dest='o_file', help='Output mpa-report file name')
    parser.add_argument('--display-header', action='store_true', dest='add_header', default=False, help='Include header in mpa-report file')
    parser.add_argument('--read_count', action='store_true', dest='use_reads', default=True, help='Use read count for output [default]')
    parser.add_argument('--percentages', action='store_false', dest='use_reads', default=True, help='Use percentages for output')
    parser.add_argument('--intermediate-ranks', action='store_true', dest='x_include', default=False, help='Include non-traditional taxonomic ranks')
    parser.add_argument('--no-intermediate-ranks', action='store_false', dest='x_include', default=False, help='Exclude non-traditional taxonomic ranks [default]')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--remove-spaces', action='store_true', dest='remove_spaces', default=True, help='Replace spaces with underscores in taxon names [default]')
    group.add_argument('--keep-spaces', action='store_false', dest='remove_spaces', default=False, help='Keep spaces in taxon names')
    args = parser.parse_args()

    kreport2mpai(args.r_file, args.o_file, add_header=args.add_header, use_reads=args.use_reads, x_include=args.x_include, remove_spaces=args.remove_spaces)

if __name__ == "__main__":
    main()
